#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sysexits.h>
#include <syslog.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <strfunc.h>

#include "lgba.h"
#include "cgi.h"

int show_attr();
int run_search();
int show_object();
int display_head();

int display_entry();
int display_count(int);
char *option_classes();
char *option_sortby();
char *option_nodes();

char *SCRIPT_NAME="index.cgi";
char *QUERY_STRING="";
sbuf *LINK=NULL;

extern char *usenode, *object;
extern slist *nodes;
extern slist *classes;
extern slist *class;

char *REMOTE_USER="unknown";
char *search=NULL;
char *sortby=NULL;
char *attr=NULL;
int  attrnum=0;
char *contenttype=NULL;
int  reverse=0;
int do_display=0;

slist *listed=NULL;

char *cf(char *cs) {
	register char *z;
	z=cfgget(cs);
	return z?z:"";
};

int tryReadConfig(char *cfname);

int
main() {
	char *p;
	slist *tl;
	int n;
	long on=1;

	// preparing getenv shadows.
	if((p=getenv("SCRIPT_NAME"))) SCRIPT_NAME=p;
	if((p=getenv("QUERY_STRING"))) QUERY_STRING=p;
	if((p=getenv("REMOTE_USER"))) REMOTE_USER=p;

	if( tryReadConfig(NULL) ) {
		printf("Content-type: text/html\n\n"
			"Failed to read config.");
		exit(EX_NOINPUT);
	};

	classes = cfgget2("classes");
	nodes   = cfgget2("nodes");
	listed  = cfgget2("listed");

	if( !classes || !nodes || !listed ) {
		fprintf(stderr, "Security violation avoided.\n");
		exit(1);
	};
	if( !nodes->count )
		fprintf(stderr, "Possible security violation: "
			"nodes list not defined.\n");

	/* Is specified node defined in config? */
	if( nodes->count ) {
		if ( (usenode = param("node")) ) {
			if(scfind(nodes, usenode) == -1)
				usenode=nodes->list[0];
		} else
			usenode=nodes->list[0];
	};

	lgba_ctl(LCTL_SET_HTMLMODE, &on);
	lgba_ctl(LCTL_SET_USE_LOCALE, &on);

	if((n=lgba_init(NULL))) {
		printf("%s\n\n", cf("content-type"));
		printf("ERROR: %s.\n", lgba_err2string(n));
		printf("%s", FOOTER);
		exit(2);
	};

	setgid(getegid());
	lgba_setuser(REMOTE_USER);

	if((n=lgba_open()) || (usenode && (n=lgba_chnode(usenode)))) {
		printf("%s\n\n", cf("content-type"));
		printf("ERROR: %s.\n", lgba_err2string(n));
		printf("%s", FOOTER);
		exit(2);
	};

	/* Sort by */
	sortby  = param("sortby");

	/* Search string */
	search=param("search");

	/* Object which we choosed */
	object=param("~");

	/* User wants to get a single attribute */
	if(object) {
		if((attr=param("attr")))
			contenttype=param("contenttype");
		if((p=param("attrnum"))) {
			attrnum=atoi(p);
			if(attrnum < 0)
				attrnum = 0;
		};
	};

	if(!contenttype)
		printf("%s\n\n", cf("content-type"));
	else
		printf("Content-type: %s\n\n", contenttype);

	fflush(stdout);

	/* Get classes we are not working with */
	class=scopy(param2("objectclass", 0));
	for(n=0; n<class->count; n++) {
		if(scfind(cfgget2("classes"), class->list[n]) == -1)
			sdel(class, n--);
	};

	if(param("reverse"))
		reverse=1;

	/* Construct a self-reference link */
	LINK=sbuf_init();
	sbuf_sprintf(LINK, "%s?node=%s", SCRIPT_NAME, usenode?usenode:"");

	/* Add search string */
	if(search)
		sbuf_sprintf(LINK, "&search=%s", url_encode(search));

	/* Add */
	if(class->count)
		sbuf_sprintf(LINK, "&objectclass=%s",
			sjoin(class, "&objectclass="));

	if((tl=getlanguageprefs())) {
		int i;
		for(i = 0; i < tl->count; i++)
			if(lgba_setlang(tl->list[i]))
				break;
	};

lgba_addfriendly("cgi(REMOTE_USER)", REMOTE_USER);
lgba_addfriendly("cgi(SCRIPT_NAME)", SCRIPT_NAME);
lgba_addfriendly("cgi(LINK)", LINK->buf);
lgba_addfriendly("cgi(search)", search?search:"");
lgba_addfriendly("cgi(node)", usenode?usenode:"");
if(sortby)
	lgba_addfriendly("cgi(sortby)", sortby);
lgba_addfriendly("cgi(option_classes)", option_classes());
lgba_addfriendly("cgi(option_sortby)", option_sortby());
lgba_addfriendly("cgi(option_nodes)", option_nodes());
lgba_addfriendly("cgi(reverse)", reverse?"checked=on":"");

	if(!contenttype)
		printf("%s", getformat(cf("search_table")));

	if(!search && !object) {
		printf("%s", FOOTER);
		exit(0);
	}

/* === */

	fflush(stdout);

	if(attr) {
		show_attr();
		lgba_close();
		exit(0);
	}

	if(object) show_object();
	else if(search) run_search();
	else printf("%s", getformat(cf("not_found")));

	lgba_close();

	printf("%s", FOOTER);

	return 0;
}

int cmp(const char *a, const char *b) {
	register int va=atoi(a);
	register int vb=atoi(b);
	if(va != vb) return va-vb;

#ifdef	HAVE_STRCOLL
	return strcoll(a, b);
#else
	return strcmp(a, b);
#endif
};

int rcmp(const char *a, const char *b) {
	register int va=atoi(a);
	register int vb=atoi(b);
	if(va != vb) return vb-va;

#ifdef	HAVE_STRCOLL
	return strcoll(b, a);
#else
	return strcmp(b, a);
#endif
}

int checkbraces(char *br) {
    int z=0;
    if(!br) return 0; 
    while(1) {
      switch(*br) { 
        case 0: return z==0?1:0;
        case '(':
            z++;
            break;
        case ')': 
            z--;
            if(z<0) return 0;
            break;
      };
      br++;
    };
};

int run_search() {
	int k;
	sbuf *filter=NULL;
	char *se=search;
	char *p;
	char **vp;
	slist *cls=class;
	slist *search_in=cfgget2("search_in");

	if(!cls->count) cls=classes;

	filter = sbuf_init();

	if(*se == '(') {
		if(!checkbraces(se)) {
			printf("%s", lgba_map_friendly("cgi::invalid_filter"));
			return -1;
		};
		sbuf_sprintf(filter, "(&%s(|", se );
		vp=cls->list;
		while(*vp)
			sbuf_sprintf(filter, "(objectclass=%s)", *vp++);
		sbuf_add(filter, "))");
	} else {;
		// simplify search
		while(*se && (*se == ' ' || *se == '*'))
			se++;
		p=se + strlen(se) - 1;
		while(p != se && (*p == ' ' || *p == '*'))
			*p--=0;
	
		// construct search filter
		if(*se) {
			p=(char *)sf_malloc(strlen(se) + 4);
			sprintf(p, "*%s*", se);
			se=p;
		} else {
			se=sf_strdup("*");
		}
	
		if(*(se+1)) {
			sbuf_add(filter, "(&(|");
			vp=search_in->list;
			while(*vp)
				sbuf_sprintf(filter, "(%s=%s)", *vp++, se);
			sbuf_add(filter, ")(|");
			vp=cls->list;
			while(*vp)
				sbuf_sprintf(filter, "(objectclass=%s)", *vp++);
			sbuf_add(filter, "))");
		} else {
			sbuf_add(filter, "(|");
			vp=cls->list;
			while(*vp)
				sbuf_sprintf(filter, "(objectclass=%s)", *vp++);
			sbuf_add(filter, ")");
		};
	};

	printf("<TITLE>Query: %s</TITLE>", filter->buf);

	fflush(stdout);

	k=lgba_search(NULL, filter->buf, sarray(listed, 0), sortby, reverse?rcmp:cmp, display_count, display_entry, 0);

	free(filter);
	free(se);

	if(!k) {
		printf("%s", getformat(cf("not_found")));
		return 0;
	} else if(k==1) {
		show_object();
		return 0;
	}

	if(do_display)
		printf("%s", cf("list_end"));

	return 0;
}

int display_count(int c) {
	if(!c) return c;
	if(c == 1) return c;
	if(c > 1) do_display=1;

	if(c > 5)
		printf("%s %d", lgba_map_friendly("cgi::users_found"), c);

	fflush(stdout);

	if(cfgget("list_start"))
		display_head();

	return c;
}

int display_entry() {
	if(do_display) format(cf("list_format"), 1);
	else object=lgba_objectname();
	return 0;
};

char **
curly(char *what, void *opt) {
	static char *vp[] = { NULL, NULL };
	static sbuf *sb = NULL;
	int z=0;

	(void)opt;

	if(sb)
		sbuf_zero(sb);
	else
		sb=sbuf_init();
	
	sbuf_sprintf(sb, "<A HREF=\"%s&sortby=%s", LINK->buf, what);
	if(sortby && (reverse ^ (z=ecq(what, sortby))))
		sbuf_add(sb, "&reverse=on");
	sbuf_sprintf(sb, "\">%s%s%s</A>",
		z?"[":"", lgba_map_friendly(what), z?"]":"");

	*vp = sb->buf;

	return vp;
};

int display_head() {
		fmt_base *fb;

		fb=format_init();
		format_metarule(fb, '`', '\'', curly);
		format_metarule(fb, '{', '}', curly);

		printf("%s", formatf(fb, cf("list_start"), NULL));
		fflush(stdout);

		format_free(fb);

        return 0;
};

char *option_classes() {
	sbuf *sb;
	char *fr;
	char **vp=sarray(classes, 0);
	char *buf;

	if(!vp) return 0;

	sb = sbuf_init();

	while(*vp) {
		fr=lgba_map_friendly(*vp);
		sbuf_sprintf(sb,
			"<INPUT TYPE=checkbox NAME=objectclass VALUE=\"%s\"%s> %s", *vp, scfind(class, *vp)!=-1?" checked=on":"", fr);

		if(*(++vp))
			sbuf_add(sb, ", ");
	}

	buf = sb->buf;
	free(sb);

	return buf;
}


char *option_sortby() {
	sbuf *sb;
	char *buf;
	int i;
	char *sby=sortby;
	char *fr, *ll;

	if(!sby) sby="";
	sb = sbuf_init();
	
	for(i=0; i<listed->count; i++) {
		ll=listed->list[i];
		fr=lgba_map_friendly(ll);
		sbuf_sprintf(sb, "<OPTION VALUE=\"%s\"%s>%s\n",
			ll, ecq(ll, sby)?" selected=on":"", fr);
	}

	buf = sb->buf;
	free(sb);
	return buf;
}

int
show_attr() {
	slist *sl;

	if(!attr || !*attr) return 0;
	if(load(object)) return 0;

	sl = getattr2(attr);

	if(attrnum < 0 || attrnum > sl->count)
		return 0;

	write(1, sl->list[attrnum], sl->lens[attrnum]);
	return 0;
}

int show_object() {
	slist *show_define_known;
	char *show_unknown=cfgget("show_unknown");
	char *show_binary=cfgget("show_binary");
	char *show_text=cfgget("show_text");
	char *at;
	char *p, *c;
	int ty;

	if(!object || load(object)) {
		printf("%s", getformat(cf("not_found")));
		return 0;
	};

	fflush(stdout);

	format(cf("show_head"), 1);

	if(!show_unknown) {
		format(cf("show_end"), 1);
		return 0;
	};
		
	if(!show_binary)
		show_binary=show_unknown;
	if(!show_text)
		show_text=show_unknown;

	show_define_known=cfgget2("show_define_known");

	fflush(stdout);

	while((at=lgba_next_attribute())) {
		if(scfind(show_define_known, at) != -1) continue;
		if(!canIread(at)) continue;

		p=show_unknown;
		c=NULL;
		ty = lgba_gettype(at);
		if(ty & TY_BINARY) {
				p=show_binary;
				c=lgba_gettypespec(at);
		};
		if(ty & TY_TEXT)
				p=show_text;

		if(!c)
			c="application/octet-stream";
		c=url_encode(c);

		p=replace(p, "$`%a'", lgba_map_friendly(at));
		p=replace(p, "%a", at);
		p=replace(p, "%c", c);
		format(p, 1);
	}


	format(cf("show_end"), 1);

	return 0;
}

