#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <lgba.h>
#include <ncurses.h>
#include <dialog.h>
#include <strfunc.h>

WINDOW *win;

#ifndef MAX_LEN
#	define MAX_LEN	2048
#endif

char on[MAX_LEN];
char cn[MAX_LEN];
char mf[MAX_LEN];
char pw[MAX_LEN];
char pwc[MAX_LEN];
char dpt[MAX_LEN];
char oph[MAX_LEN];
char hph[MAX_LEN];

char *trygetuser();
void clearnexit();
int iterate(int);
int trytocreate();
void *makedepts();
void *maketitles();
int drawback();

int deptscount=0;
int titlescount=0;

dialogMenuItem *dptr=NULL;
dialogMenuItem *tptr=NULL;

int p_checked(struct _dmenu_item *self);
int p_fire(struct _dmenu_item *self);
void p_selected(struct _dmenu_item *self, int is_selected);

int main() {
	char *ep;

	win=initscr();
	cbreak(); noecho();
	init_dialog();

	title_attr=dialog_attr;

	if(lgba_init(NULL) || lgba_open()) {
		dialog_mesgbox("Error", "LGBA Connection error!", 6, 20);
		clearnexit();
	};

	*on=*cn=0;

	while((ep=trygetuser()))
		dialog_mesgbox("Error", ep, 6, strlen(ep)>12?strlen(ep)+2:15);

	drawback();

	while(iterate(0)) {
		DialogY=10;
		if(!dialog_yesno("Save", "Save this user?", 5, 10))
			trytocreate();
		drawback();
	};

	dialog_update(); endwin(); exit(0);

	dialog_update();

	clearnexit();
};

char *trygetuser() {
	if(dialog_inputbox("Add", "Add new user", 8, 30, on)) clearnexit();

	if(!*on) return "Empty user name";

	if(load(on) != LGBA_NOTFOUND) {
		if(lgba_errno == LGBA_NOERR) return "Already exists";
		return lgba_err2string(lgba_errno);
	};

	return NULL;
};

void clearnexit() {
	end_dialog();
	clear();
	refresh();
	endwin();
	exit(0);
};

int iterate(int z) {
	int ec;

	if(!dptr) dptr=makedepts();
	if(!tptr) tptr=maketitles();

	switch(z) {
		case 0:
			ec=line_edit(win, 5, 7, 128, 18, 0, 1, cn, 0);
			if(ec == 259) return iterate(0);
		case 1:
			ec=line_edit(win, 9, 7, 128, 18, 0, 1, mf, 0);
			if(ec == 259) return iterate(0);
		case 2:
			ec=line_edit(win, 13, 7, 128, 18, 0, 1, pw, 1);
			if(ec == 259) return iterate(1);
		case 3:
			ec=line_edit(win, 15, 7, 128, 18, 0, 1, pwc, 1);
			if(ec == 259) return iterate(2);
			while(!eq(pw, pwc)) {
				mvwaddstr(win, 14, 7, "try again...");
				ec=line_edit(win, 15, 7, 128, 18, 0, 1, pwc, 1);
				if(ec == 259) return iterate(2);
			};
		case 4:
			ec=line_edit(win, 5, 28, 128, 18, 0, 1, oph, 0);
			if(ec == 259) return iterate(3);
		case 5:
			ec=line_edit(win, 9, 28, 128, 18, 0, 1, hph, 0);
			if(ec == 259) return iterate(4);
		case 10:
			if(dptr) {
				ec=-1; while(++ec < deptscount)
					dptr[ec].mark=(dptr[ec].mark=='X')?' ':'X';
				if(dialog_checklist("Department", "Choose a department",
				18, 65, 10, -deptscount, dptr, NULL)) {
					drawback();
					return iterate(5);
				};
			};
		case 11:
			if(tptr) {
				ec=-1; while(++ec < titlescount)
					tptr[ec].mark=(tptr[ec].mark=='X')?' ':'X';
				if(dialog_checklist("Title", "Choose one or more titles",
				18, 55, 10, -titlescount, tptr, NULL)) return iterate(10);
			};
	};

	return 1;
}

int trytocreate() {
	if(create(on, NULL)) {
		sleep(1);
		dialog_mesgbox("Create error", lgba_err2string(lgba_errno), 6, 20);
		return 0;
	};

	if(*cn) setattr("cn", cn);
	if(*mf) {
		slist *sl = split(mf, ", \t", 0);
		if(sl->count) setattr2("mail", mkarray(sl, 0));
		sfree(sl);
	};

	if(*oph) setattr("officephone", oph);
	if(*hph) setattr("homephone", hph);

	if(*pw) setattr("userPassword", pw);

	if(dptr) {
		slist *sl=sinit();
		int z=-1;
		while(++z < deptscount)
			if(dptr[z].mark=='X') sadd(sl, dptr[z].prompt);
		if(sl->count)
			setattr2("Department", mkarray(sl, 0));
		sfree(sl);
	};

	if(tptr) {
		slist *sl=sinit();
		int z=-1;
		while(++z < titlescount)
			if(tptr[z].mark=='X') sadd(sl, tptr[z].prompt);
		if(sl->count)
			setattr2("Title", mkarray(sl, 0));
		sfree(sl);
	};

	if(lgba_save(1) || load(on)) {
		dialog_mesgbox("Save error", lgba_err2string(lgba_errno), 6, 20);
		clearnexit();
	};

dialog_mesgbox("All clean!",
getformat("\tUser created!

Object:     \t${~}
Common name:\t${cn}
${mail?Mail\\:       \t${mail+\n\t\t}\n}${department?Department\\: \t${department+\n\t\t}\n}${title?Title\\:      \t${title+\n\t\t}\n}${uidnumber?UID\\:        \t${uidnumber}}
"),
	15, 40);

	end_dialog();
	endwin();
	exit(0);

	clearnexit();
};

void *makedepts() {
	dialogMenuItem *iptr;
	static slist *a=NULL;
	static slist *b=NULL;
	char *ts;
	int k=0;

	if(a) sclear(a);
	if(b) sclear(b);

	if(!(ts=lgba_gettypespec("department"))) return NULL;

	if(lgba_readavfile(ts, &a, &b)) return NULL;

	if(!(deptscount=a->count)) return NULL;

	iptr=(dialogMenuItem *)sf_calloc(a->count, sizeof(dialogMenuItem));

	while(k < a->count) {
		iptr[k].prompt=*(a->list[k])?a->list[k]:" ";
		iptr[k].title=*(b->list[k])?b->list[k]:" ";
		if(a->maxlen + strlen(iptr[k].title) > 48)
			strcpy(&iptr[k].title[48 - a->maxlen], "...");
		iptr[k].checked=p_checked;
		iptr[k].fire=p_fire;
		iptr[k].selected=p_selected;
		iptr[k].data=sf_strdup("   ");
		iptr[k].lbra='[';
		iptr[k].mark=0;
		iptr[k].rbra=']';
		k++;
	};

	return iptr;
};

void *maketitles() {
	dialogMenuItem *iptr;
	static slist *a=NULL;
	static slist *b=NULL;
	char *ts;
	int k=0;

	if(a) sclear(a);
	if(b) sclear(b);

	if(!(ts=lgba_gettypespec("title"))) return NULL;

	if(lgba_readavfile(ts, &a, &b)) return NULL;

	if(!(titlescount=a->count)) return NULL;

	iptr=(dialogMenuItem *)sf_calloc(a->count, sizeof(dialogMenuItem));

	while(k < a->count) {
		iptr[k].prompt=*(a->list[k])?a->list[k]:" ";
		iptr[k].title=*(b->list[k])?b->list[k]:" ";
		if(a->maxlen + strlen(iptr[k].title) > 38)
			strcpy(&iptr[k].title[34 - a->maxlen], "...");
		iptr[k].checked=p_checked;
		iptr[k].fire=p_fire;
		iptr[k].selected=p_selected;
		iptr[k].data=sf_strdup("   ");
		iptr[k].lbra='[';
		iptr[k].mark=0;
		iptr[k].rbra=']';
		k++;
	};

	return iptr;
};

int p_checked(struct _dmenu_item *self) {
	if(!self->mark) {
		self->mark=' ';
		return 0;
	};
	self->mark=(self->mark==' ')?'X':' ';
	return (self->mark==' ')?0:1;
};

int p_fire(struct _dmenu_item *self) {
	return 1;
};

void p_selected(struct _dmenu_item *self, int iss) {
};


int drawback() {
	char *z, *p;

	dialog_clear_norefresh();
	draw_box(win, 2, 5, 20, 70, dialog_attr, border_attr);
	draw_shadow(win, 2, 5, 20, 70);

	attrset(dialog_attr);
	mvwaddstr(win, 2, 30, " Adding user: ");
	mvwaddstr(win, 2, 44, on);
	mvwaddstr(win, 2, 44+strlen(on), " ");

	draw_box(win, 4, 6, 3, 20, dialog_attr, border_attr);
	attrset(dialog_attr);
	mvwaddstr(win, 3, 7, "Common name:");
	mvwaddstr(win, 5, 7, cn);

	draw_box(win, 8, 6, 3, 20, dialog_attr, border_attr);
	attrset(dialog_attr);
	mvwaddstr(win, 7, 7, "Mail (forward):");
	mvwaddstr(win, 9, 7, mf);

	draw_box(win, 12, 6, 5, 20, dialog_attr, border_attr);
	attrset(dialog_attr);
	mvwaddstr(win, 11, 7, "Password:");
	mvwaddstr(win, 14, 7, "confirm...");
	p=z=sf_strdup(pw); while(*p) *p++='*';
	mvwaddstr(win, 13, 7, z);
	free(z);
	p=z=sf_strdup(pwc); while(*p) *p++='*';
	mvwaddstr(win, 15, 7, z);
	free(z);

	draw_box(win, 4, 27, 3, 20, dialog_attr, border_attr);
	attrset(dialog_attr);
	mvwaddstr(win, 3, 28, "Office phone:");
	mvwaddstr(win, 5, 28, cn);

	draw_box(win, 8, 27, 3, 20, dialog_attr, border_attr);
	attrset(dialog_attr);
	mvwaddstr(win, 7, 28, "Home phone:");
	mvwaddstr(win, 9, 28, mf);

	dialog_update();

	return 0;
};

