#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <strfunc.h>
#include "lgba.h"


char *link1="department";	/* User exists in the specified network */
char *link2="department";	/* Department consists of those networks */
char *left="webresourceleft";	/* Counter */
char *expires="expires";	/* Expires, in seconds */
char *userpwd="userpassword";	/* Expires, in seconds */
char *ocls="ProxyAccess";	/* This objectclass must be present */
char *node=NULL;
char *dnode="Networks";

int sfunc(int);
void usage();

int main(int ac, char **av) {
	char buf[16384];
	int t;
	slist *sl;
	char *dnodedn=NULL;
	slist *avs;
	char *p;
	char *flt=NULL;
	char **nwks;
	long off = 0;

	extern char *optarg;
	extern int optind;
	int ch;

	if(ac == 1) usage();

	while ((ch = getopt(ac, av, "l:e:o:n:d:1:2:h?")) != -1)
	switch (ch) {
		case 'l': left=optarg; break;
		case 'e': expires=optarg; break;
		case 'o': ocls=optarg; break;
		case 'n': node=optarg; break;
		case 'd': dnode=optarg; break;
		case '1': link1=optarg; break;
		case '2': link2=optarg; break;
		case 'h':
		case '?':
		default:
			usage();
	}
	ac -= optind;
	av += optind;


	setgid(getegid());

	if(lgba_init(NULL) || lgba_open()) {
		fprintf(stderr, "Can't init LGBA\n");
		exit(0);
	};

	if(lgba_setuser("root")) {
		fprintf(stderr, "Can't set LGBA user to 'root'\n");
		exit(0);
	};

	lgba_ctl(LCTL_SET_OVERRIDEMODE, &off);

	if(dnode) {
		if(lgba_chnode(dnode))
		fprintf(stderr, "Warning: Can't find node '%s'\n", dnode);
		else
		dnodedn=sf_strdup(lgba_current_dn());
	};

	if(lgba_chnode(node)) {
		fprintf(stderr, "Can't select default node\n");
		exit(0);
	};


	avs=sinit();
	sl=sinit();

	sadd(avs, userpwd);
	sadd(avs, left);
	sadd(avs, expires);
	sadd(avs, link1);
	sadd(avs, link2);

	lgba_setattrlist(sarray(avs, 0));

	while(fgets(buf, sizeof(buf), stdin)) {
		sclear(sl);
		splitf(sl, buf, NULL, 0);
		if(sl->count != 2)
			{ write(1, "ERR\n", 4); continue; };

		if(load(sl->list[0]))
			{ write(1, "ERR\n", 4); continue; };

		if(cfind(lgba_objectclass(), ocls) == -1)
			{ write(1, "ERR\n", 4); continue; };

		if(lgba_verifypwd(sl->list[1]))
			{ write(1, "ERR\n", 4); continue; };

		if(atoi(getattr(left)) <= 0)
			{ write(1, "ERR\n", 4); continue; };

		if((p=getattr(expires)) && (t=atoi(p)) != 0)
			if(difftime(time(NULL), t) <= 0)
				{ write(1, "ERR\n", 4); continue; };

		if(!dnodedn || !(nwks=sarray(getattr2(link1), 0)) || !*nwks)
			{ write(1, "OK\n", 3); continue; };

		nwks--;
		t=0;
		while(*++nwks) {

		flt=(char *)sf_realloc(flt,
			strlen(*nwks) + strlen(link2) + 10);
		sprintf(flt, "(%s=%s)", link2, *nwks);
		if(lgba_search(dnodedn, flt, NULL, NULL, NULL, NULL, sfunc, 1))
			{ t=1; break; };

		};

		/* Department isn't found, so
		treat user as privileged */

		if(!t) { write(1, "OK\n", 3); continue; };

		if((p=getattr(left)) && *p)
			if(atoi(p) <= 0)
				{ write(1, "ERR\n", 4); continue; };

		if((p=getattr(expires)) && (t=atoi(p)) != 0)
			if(difftime(time(NULL), t) >= 0)
				{ write(1, "ERR\n", 4); continue; };

		write(1, "OK\n", 3);
	}

	return 0;
}

int sfunc(int n) { (void)n; return 1; };


void usage() {
	fprintf(stderr, "Squid Authenticator (c) vlm@spelio.net.ru, 1999, 2000.
Usage: squid_auth [-l <resourceattr>] [-e <expireattr>] [-o <requiresobjectclass>] [-n <node>] [-d <dependentnode>] [-1 <linkattributeinnode>] [-2 <linkattributeindependentnode>]\n");
	fprintf(stderr, "Defaults:\n");
	fprintf(stderr, "\t-l: %s\n", left);
	fprintf(stderr, "\t-e: %s\n", expires);
	fprintf(stderr, "\t-o: %s\n", ocls);
	fprintf(stderr, "\t-1: %s\n", link1);
	fprintf(stderr, "\t-2: %s\n", link2);
	fprintf(stderr, "\t-n: %s%s%s\n", node?"":"[", node, node?"":"] (default node)");
	fprintf(stderr, "\t-d: %s%s%s\n", dnode?"":"[", dnode, dnode?"":"] (default node)");
	exit(0);
};


