#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sysexits.h>
#include <ctype.h>

#include <lgba.h>
#include <strfunc.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef FW_DEL_PROG

#ifdef	HAVE_NETINET_IP_FW_H
#include <netinet/ip_fw.h>
#else
#error "Direct firewall access not implemented at this platform."
#error "Define -DFW_DEL_PROG in the Makefile."

#endif

#endif	/* FW_DEL_PROG */

int realwork=0;

int sux(const char *);
void display(FILE *);
void store(char *, char *, int, char *, char *);
void delrule(int);

char smallbuf[128];	/* small buffer */
int decreasenetwork(int sizek, int try);
int sfunc(int);

static char *atl[]={
	"webresourceleft", "webresourcequota", "webresourcemax",
	"webresourcespent",
	"networks", "network", "firewallrule", NULL };

typedef struct _hostz {
	char *login;
	struct in_addr ip;
	char local;
	int network;
	size_t allsize, allnum;
	struct _hostz *next;
} hostz;

hostz hosts;

void sigint(int sig) {
	sig=realwork;
	realwork=0;
	display(stderr);
	realwork=sig;
};

ino_t inode;
long fseek_to;

ino_t finode(int fd) {
	struct stat sb;
	fstat(fd, &sb);
	return sb.st_ino;
}

long fsize(int fd) {
	struct stat sb;
	fstat(fd, &sb);
	return sb.st_size;
}

extern void * node_head;

void usage() {
	fprintf(stderr, "Sqlimit (c) Lev Walkin, <vlm@spelio.net.ru>, 2000\n"
	"Usage: sqlimit [-r] [-l <local>] [-i <ignore>] <squid_access.log>\n"
	"\t-r:\tDo real work (test mode otherwise)\n"
	"\t-l:\tFile with local domains\n"
	"\t-i:\tFile with URLs to ignore\n"
	);
	exit(EX_USAGE);
}

slist *readfile(char *fname);

slist *localdomains = NULL;
slist *ignoreurls = NULL;
int ignoreurls_f(char *what);
int localdomains_f(char *what);

int main(int ac, char **av) {
	FILE *fd;
	char *posext;
	int  posfd;
	sbuf *sb;
	slist *sl;
	char *p;
	char *ipaddr;
	char *status;
	char *url;
	char *who;
	char *op;
	char *acc;
	int size;
	int c;

	if(ac < 1)
		usage();

	while((c=getopt(ac, av, "rl:i:")) != -1)
	switch(c) {
		case 'r':
			realwork = 1;
			break;
		case 'l':
			localdomains=readfile(optarg);
			break;
		case 'i':
			ignoreurls=readfile(optarg);
			break;
		default:
			usage();
	};

	ac -= optind;
	av += optind;

	if(ac != 1)
		usage();

	signal(SIGINT, sigint);

	fd=fopen(*av, "r");
	if(!fd) {
		fprintf(stderr, "Can't open file: %s\n", *av);
		exit(EX_NOINPUT);
	};

	if(realwork) {
		if(lgba_init(NULL) || lgba_open()) {
			lgba_perror("LGBA Init error");
			exit(2);
		};
		lgba_setattrlist(atl);
	};

	/* Determine the file position indicator extension */
	if(realwork)
		posext="sqlimit.pos";
	else
		posext="sqlimit.pos.test";

	posfd = initFILEpos(*av, fd, posext);
	if(posfd == -1) {
		fprintf(stderr, "Can't open position file cache.\n");
		exit(EX_NOINPUT);
	};

	memset(&hosts, 0, sizeof(hosts));

	sl = sinit();
	sb = sbuf_init();

	while((sbuf_zero(sb) == 0) && (p = sbuf_fgets(sb, fd))) {
		sclear(sl);
		if(splitf(sl, p, NULL, 0) != 10) {
			fprintf(stderr, "[Parse error: split by %d]\n",
				splitf(sl, p, NULL, 0));
			continue;
		};
		ipaddr = sl->list[2];
		status = sl->list[3];
		size = atoi(sl->list[4]);
		op = sl->list[5];
		url = sl->list[6];
		who = sl->list[7];
		acc = sl->list[8];
		if(!strncmp(acc, "NONE", 4)) continue;
		if(ignoreurls && ignoreurls_f(url))
			continue;
		store(ipaddr, url, size, who, acc);
	};
	adjFILEpos(fd, posfd, 1);
	fclose (fd);
	close(posfd);
	display(stdout);

	sfree(sl);
	sbuf_free(sb);

	if(realwork)
		lgba_close();

	return 0;
}

void store(char *ipaddr, char *url, int size, char *login, char *acc) {
	struct in_addr ip;
	char local='F';
	int network=0;
	hostz *h, *oh;
	unsigned char ipn[4];

	if(!strcmp(login, "-") || !strcmp(login, "anonymous")) login=NULL;

	inet_aton(ipaddr, &ip);
	*((unsigned long *)ipn)=ip.s_addr;

	if(!login && ipn[2] != 1 && ipn[2] != 50 && ipn[2] != 9) {
		ipn[3]=0;
		ip.s_addr=*((unsigned long *)ipn);
		network=1;
	};

	if(localdomains && localdomains_f(url))
			local = 'L';

	h=&hosts;
	oh=&hosts;
	while(h) {
		if((!login && !h->login && h->ip.s_addr == ip.s_addr && local==h->local) || (login && h->login && !strcmp(login, h->login))) {
			h->allsize+=size;
			h->allnum++;
			break;
		};
		oh=h;
		h=h->next;
	}

	if(!h) {
		oh->next=(hostz *)sf_calloc(1, sizeof(hostz));
		oh=oh->next;
		oh->ip=ip;
		if(login) oh->login=sf_strdup(login);
		oh->network=network;
		oh->local=local;
		oh->allsize=size;
		oh->allnum=1;
	};

};

int sux(const char *msg) {
	fprintf(stderr, "%s\n", msg);
	exit(0);
}

void display(FILE *fp) {
	hostz *h=&hosts;
	char *net;
	char **vp;
	int z;
	int SIZE;
	sbuf *sb;

	while(h->next) {
		h=h->next;
		if(!h->login) continue;
		if(h->network) net="/24";
		else net="/32";
		fprintf(fp, "%c %s \t\t%u \t(%d)\n", h->local, h->login, h->allsize, h->allnum);

		if(!realwork) continue;
		if(h->local == 'L') continue;
		if(load(h->login)) continue;
		SIZE=(int)((h->allsize+512)/1024);
		z = atoi(getattr("webresourceleft"));
		sprintf(smallbuf, "=%d", z - SIZE);
		setattr("webresourceleft", smallbuf);
		fprintf(stderr, "%s: %d-%d => %s\n",
			lgba_objectname(), z, SIZE, smallbuf);
		setattr("webresourcespent",
			genstr(atoi(getattr("webresourcespent")) + SIZE));
		if(!(vp=sarray(getattr2("network"), 0))) continue;

		lgba_chnode("OU"); 	/* Organizational units */
		z=0;
		vp--;
		while(*(++vp) && !z)
			if(!load(*vp))
				z |= decreasenetwork(SIZE, 1);
		if(!z && !load(*vp))
				decreasenetwork(SIZE, 0);
		lgba_chnode(NULL);
	};
	fprintf(fp, "Departments:\n");
	if(realwork) {
		lgba_chnode("OU");
		if(lgba_errno) {
			lgba_perror("OU");
			realwork=0;
		};
	};

	h=&hosts;
	sb = sbuf_init();

	while(h->next) {
		h=h->next;
		if(h->login) continue;
		if(h->network) net="/24";
		else net="/32";
		fprintf(fp, "%c %s%s \t%u \t(%d)\n", h->local, inet_ntoa(h->ip), net, h->allsize, h->allnum);
		if(!realwork) continue;
		if(h->local == 'L') continue;

		sbuf_zero(sb);
		sbuf_sprintf(sb, "(networks=%s%s)",
			inet_ntoa(h->ip), h->network?"/24":"/32");
		if(!lgba_search(NULL, sb->buf, atl, NULL, NULL, NULL, sfunc, 0)) {
			fprintf(stderr, "Decrease failed for %s\n", inet_ntoa(h->ip));
			continue;
		};
		SIZE=(int)((h->allsize+512)/1024);
		if(!decreasenetwork(SIZE, 0))
			fprintf(stderr, "Decrease failed for %s\n", inet_ntoa(h->ip));
	};

	sbuf_free(sb);

};

int decreasenetwork(int sizek, int try) {
	int newsize;
	char *s;
	int z;

	if(!lgba_objectdn()) return 0;

	newsize=(z=atoi(getattr("webresourceleft"))) - sizek;
	if(try && newsize <= 0) return 0;

	sprintf(smallbuf, "=%d", newsize);
	setattr("webresourceleft", smallbuf);
	fprintf(stderr, "%s: \t%d - %d => \t%d\n",
		lgba_objectname(), z, sizek, newsize);
	setattr("webresourcespent",
		genstr(atoi(getattr("webresourcespent")) + sizek));

	if(newsize <= 0) {
		s=getattr("firewallrule");
		if(!*s || !atoi(s)) {
			fprintf(stderr, "Unknown firewall rule!\n");
			return 0;
		};
		delrule(atoi(s));
	};

	return 1;
};

int sfunc(int z) {
	return 1;
};

void delrule(int num) {
#ifdef FW_DEL_PROG
	snprintf(smallbuf, sizeof(smallbuf), FW_DEL_PROG, num);
	fprintf(stderr, "%s\n", smallbuf);
	system(smallbuf);
	return;
#endif

#ifdef	HAVE_NETINET_IP_FW_H
	struct ip_fw rule;
	static int sock=-1;

	if((sock=-1)) {
		sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
		if(sock < 0) {
			fprintf(stderr, "Firewall socket error.\n");
			exit(2);
		};
	};

	memset(&rule, 0, sizeof(rule));
	rule.fw_number=num;

	if(setsockopt(sock, IPPROTO_IP, IP_FW_DEL, &rule, sizeof(rule)))
	fprintf(stderr, "Rule %d: error deleting firewall entry.\n", num);
	else fprintf(stderr, "Firewall entry %d removed.\n", num);

#endif
};


slist *
readfile(char *fname) {
	FILE *fp;
	slist *sl=NULL;
	sbuf *sb;
	char *p;

	fp = fopen(fname, "r");
	if(!fp) {
		fprintf(stderr, "Can't open file %s\n", fname);
		exit(EX_NOINPUT);
	};

	sl = sinit();
	sb = sbuf_init();

	while((sbuf_zero(sb) == 0) && (p = sbuf_fgets(sb, fp))) {
		p[sb->len-1] = 0;	/* Strip '\n' */
		sadd(sl, sb->buf);
	};

	fclose(fp);

	sbuf_free(sb);

	if(!sl->count) {
		sfree(sl);
		fprintf(stderr, "File %s contains no entries.", fname);
		sl=NULL;
	};

	return sl;
};


int
localdomains_f(char *url) {
	register int i;
	register char *p;
	register size_t dml;

	if(!localdomains || !url) return 0;

	p = url;
	while(*p != '/') p++;
	if(*++p != '/') return 0;
	if(*p++ != '/') return 0;
	if(!*p) return 0;
	dml = 0;
	while(*p != '/') { dml++; p++; };
	*p = 0;

	for(i = 0; i < localdomains->count; i++) {
		if(localdomains->lens[i] > dml)
			continue;
		if(ecq(p-localdomains->lens[i], localdomains->list[i]))
			return 1;
	};

	return 0;
};

int
ignoreurls_f(char *url) {
	register int i;

	if(!ignoreurls) return 0;

	for(i = 0; i < ignoreurls->count; i++) {
		if(!strncasecmp(ignoreurls->list[i], url, ignoreurls->lens[i]))
			return 1;
	};

	return 0;
};



