#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <err.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_NETINET_IP_FW_H
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip_fw.h>
#include <netinet/tcp.h>
#endif

#include <strfunc.h>
#include <lgba.h>

char *fwrule="firewallrule";
char *left="webresourceleft";
char *networks="networks";
char *proxynetworks="proxynetworks";
char *forward_ip="forward_ip";
char *forward_port="forward_port";

char *defaultnets[]={ "192.168.0.0/16", NULL };
char *defaultfwdtoip="192.168.1.1";
char *defaultfwdtoport="4000";
char *node="OU";

char **controlnets=NULL;
char *fwdtoip=NULL;
char *fwdtoport=NULL;

char *filter_template = NULL;
char *tail_template = NULL;

void usage();
void init();
void flush();
void bad();
void good();

/* Iterate through all available firewall rules */
void iterate(int (*func)(int));

/* Add specific firewall rule */
int addrule(int ruleno, char **);

/* Remove specific firewall rule */
int rmrule(int ruleno);

#ifdef	HAVE_NETINET_IP_FW_H
int redirector();
#endif

int quiet=0;
int sock;

int
main(int ac, char **av) {
	extern char *optarg;
	extern int optind;
	slist *avl;
	int ch;

#ifndef	HAVE_NETINET_IP_FW_H
	fprintf(stderr,
	"Direct access to the FreeBSD IP Firewall is not compiled-in.\n"
	"Probably, there is another way to modify the firewall\n"
	"table directly, but I don't know how to handle it now.\n"
	"Entering the demo mode.\n\n"
	);
#endif

	while ((ch = getopt(ac, av, "qn:h?f:t:")) != -1)
	switch (ch) {
		case 'q': quiet=1; break;
		case 'n': node=optarg; break;
		case 'f': filter_template = optarg; break;
		case 't': tail_template = optarg; break;

		case 'h':
		case '?':
		default:
			usage();
	}

	ac -= optind;
	av += optind;

	if(ac < 1) usage();

	if(lgba_init(NULL) || lgba_open()) {
		lgba_perror("LGBA init error");
		exit(2);
	};

	if(lgba_chnode(node)) {
		fprintf(stderr, "Cannot change working node to '%s'.\n", node);
		exit(2);
	};

	avl=sinit();

	sadd(avl, "cn");
	sadd(avl, "objectclass");
	sadd(avl, fwrule);
	sadd(avl, left);
	sadd(avl, networks);
	sadd(avl, proxynetworks);
	sadd(avl, forward_ip);
	sadd(avl, forward_port);

	lgba_setattrlist(sarray(avl, 0));
	sfree(avl);

#ifdef	HAVE_NETINET_IP_FW_H
	if((sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1) {
		fprintf(stderr, "Firewall socket open error.\n");
		exit(2);
	};
#endif

	if(ecq2(*av, "init")) init(); 
	else if(ecq2(*av, "flush")) flush();
	else if(ecq2(*av, "bad")) bad();
	else if(ecq2(*av, "good")) good();
	else usage();

	close(sock);
	lgba_close();

	return 0;
};

void usage() {
	fprintf(stderr, "Usage:
sqlm [-q] [-n <node>] [-f <filter_template>] [-t <tail_template>] <word>
	Where word is one from the following:
	init	- install an appropriate firewall table into the kernel
	flush	- clear the whole table (range from 10000 up to 20000)
	bad	- remove entries of objects with account debts
	good	- install entries for object with no account debts
");
	exit(EX_USAGE);
};

typedef struct chain_ {
	int rule;
	int left;
	char *name;
	char **networks;
	struct chain_ *next;
} chain;

chain *ch=NULL;
chain *ch_last=NULL;
int chainsize=0;

int addobj(int n) {
	register int z;
	register chain *c;
	register slist *nets;

	z=atoi(getattr(fwrule));
	if(z < 10000 || z >= 20000) return 0;

	if((!(nets=getattr2(proxynetworks)) &&
	!(nets=getattr2(networks))) || !nets->count) return 0;

	c=(chain *)sf_malloc(sizeof(chain));
	c->networks = mkarray(nets, 0);
	c->rule=z;
	c->left=atoi(getattr(left));
	c->name=sf_strdup(getformat("${~} (${cn})"));
	c->next=NULL;

	if(!ch || !ch_last)
		ch=ch_last=c;
	else
		ch_last=ch_last->next=c;

	chainsize++;

	return 0;
};

int
getobjects() {
	int z;

	if(ch) return chainsize;

	z=lgba_search(NULL, NULL, NULL, NULL, NULL, NULL, addobj, 0);
	if(!quiet)
	if((z || chainsize) && z != chainsize) {
		fprintf(stderr, "Warning: %d objects found, but %d actually added.\n", z, chainsize);
	};

	controlnets = NULL;
	/*
	if(load("control") == 0) {
		char **nets=getattr2(networks);
		if(nets && *nets)
			copy_values(nets, &controlnets);
		fwdtoip=getattr("forward_ip");
		fwdtoip=getattr("forward_port");
	}
	*/

	if(!controlnets || !*controlnets) {
		copy_values(defaultnets, &controlnets);
	};
	if(!fwdtoip || !*fwdtoip) fwdtoip=defaultfwdtoip;
	if(!fwdtoport || !*fwdtoport) fwdtoport=defaultfwdtoport;

	return chainsize;
};

void
init() {
	chain *c;
	int i;

	if(quiet) getobjects();
	else if(!getobjects())
		fprintf(stderr, "Warning: No objects in node '%s'\n", node);
	else
		fprintf(stderr, "Objects found: %d. Installing IPFW table...\n",
		chainsize);

	if(!quiet) fprintf(stderr, "Flushing previous IPFW table...\n");

	iterate(rmrule);

	if(!quiet) fprintf(stderr, "Installing permission rules...\n");

	c=ch;
	while(c) {
		if(c->left <= 0) {
			if(!quiet)
				fprintf(stderr, "[%d] ", c->rule);
			c=c->next;
			continue;
		};
		i=addrule(c->rule, c->networks);
		if(!quiet) {
			fprintf(stderr, "%s%d ", i?"!":"", c->rule);
			fflush(stderr);
		};
		c=c->next;
	};

#ifdef	HAVE_NETINET_IP_FW_H
	if(!quiet) {
		fprintf(stderr, "\nInstalling redirector at 19999: ");
		if(redirector()) fprintf(stderr, "FAIL!\n");
		else fprintf(stderr, "OK.\n");
	} else redirector();
#else
	if(!quiet)
		fprintf(stderr, "\nPlease install redirector manually.");
#endif

	if(!quiet) fprintf(stderr, "Done.\n");
};

void flush() {
	if(!quiet) fprintf(stderr, "Flushing IPFW table...\n");
	iterate(rmrule);
	if(!quiet) fprintf(stderr, "Done.\n");
};


void bad() {
	chain *c;
	register int i;
	register int z;

	if(quiet) getobjects();
	else if(!getobjects()) {
		fprintf(stderr, "No entries with debts found\n");
		exit(0);
	} else fprintf(stderr, "Removing entries with debts:\n");

	c=ch; i=0;
	while(c) {
		if(c->left <= 0) {
			if((z=quiet))
				rmrule(c->rule);
			else {
				quiet=1;
				if(!rmrule(c->rule))
					fprintf(stderr, "\n%d: %s",
					++i, c->name);
				quiet=z;
			};
		};
		c=c->next;
	};

	if(!quiet) fprintf(stderr, "%sDone.\n", i?"\n\n":"Nothing to remove.\n");
};

void good() {
	chain *c;
	register int i, z, w;

	if(quiet) getobjects();
	else if(!getobjects()) {
		fprintf(stderr, "No entries without debts found\n");
	} else fprintf(stderr, "Adding entries without debts:\n");

	c=ch; i=0;
	while(c) {
		if(c->left > 0) {
			z=quiet;
			quiet=1;
			w=rmrule(c->rule);
			addrule(c->rule, c->networks);
			quiet=z;
			if(!quiet && w)
			fprintf(stderr, "\n%d: %s", ++i, c->name);
		};
		c=c->next;
	};

	if(!quiet) fprintf(stderr, "%sDone.\n", i?"\n\n":"Nothing to add.\n");
};


#ifdef	HAVE_NETINET_IP_FW_H
int rmrule(int ruleno) {
	struct ip_fw rule;
	register int result;

	memset(&rule, 0, sizeof(rule));
	rule.fw_number=ruleno;

	result=setsockopt(sock, IPPROTO_IP, IP_FW_DEL, &rule, sizeof(rule));

	if(quiet == 0) {
	    if(result) fprintf(stderr, "!%d ", ruleno);
	    else fprintf(stderr, "%d ", ruleno);
	};

	return result;
};
#else
int
rmrule(int ruleno) {
	printf("D %d\n", ruleno);
	return 0;
};
#endif


#ifdef	HAVE_NETINET_IP_FW_H
void
iterate(int (*func)(int)) {
        struct ip_fw *rules;
        void *data = NULL;
        int n, num = 0;
        int nbytes;

        /* get rules or pipes from kernel, resizing array as necessary */
        {
                const int unit = sizeof(*rules);
                const int ocmd = IP_FW_GET;
                int nalloc = 0;
                while (num >= nalloc) {
                        nalloc = nalloc * 2 + 200;
                        nbytes = nalloc * unit;
                        if ((data = realloc(data, nbytes)) == NULL)
                                err(EX_OSERR, "realloc");
                        if (getsockopt(sock, IPPROTO_IP, ocmd, data, &nbytes)
				< 0)
                                err(EX_OSERR, "getsockopt(IP_%s_GET)", "FW");
                        num = nbytes / unit;
                }
        }
        rules = (struct ip_fw *) data;
        /* determine num more accurately */
        num = 0;
        while (rules[num].fw_number < 65535)
            num++ ;
        num++ ; /* counting starts from 0 ... */
	/* display all rules */
	nbytes=0;
	for (n = 0; n < num; n++) {
		struct ip_fw *const r = &rules[n];                      
		if(r->fw_number >= 10000 && r->fw_number < 20000) {
			func(r->fw_number);
			nbytes++;
		};
	}
        free(data);
	if(!quiet && nbytes) fprintf(stderr, "\n");
};
#else
void
iterate(int (*func)(int)) {
	printf("R removing the whole firewall table.\n");
};
#endif

#ifdef	HAVE_NETINET_IP_FW_H

int
lookup_host (char *host, struct in_addr *ipaddr) {
        struct hostent *he;
	if (!inet_aton(host, ipaddr)) {
		if ((he = gethostbyname(host)) == NULL)
			return(-1);
		*ipaddr = *(struct in_addr *)he->h_addr_list[0];
	}
        return(0);
};

int fill_ip(struct in_addr *ipno, struct in_addr *mask, char *str) {
        char *p, md = 0;
	p = strchr(str, '/');
	if (!p) p = strchr(str, ':');
	if (p) { md = *p; *p++ = '\0'; }
	if (lookup_host(str, ipno)) return 1;
	if(!mask) return 0;
	switch (md) {
		case ':':
			if (!inet_aton(p, mask)) return 2;
		case '/':
			if (atoi(p) == 0) {
				mask->s_addr = 0;
			} else if (atoi(p) > 32)
				return 3;
			else
				mask->s_addr =
				htonl(~0 << (32 - atoi(p)));
			break;
		default:
			mask->s_addr = htonl(~0);
			break;
	}
	ipno->s_addr &= mask->s_addr;
	return 0;
};

int
addrule(int ruleno, char **nets) {
	struct ip_fw rule;
	int ret = 0;

	if(ruleno < 10000 || ruleno >= 20000 || !nets || !*nets) return 1;

	if(filter_template) {
		static sbuf *sb = NULL;

		if(!sb)
			sb = sbuf_init();

		for(; *nets; nets++) {
			sbuf_zero(sb);
			sbuf_sprintf(sb,
				"/sbin/ipfw add %d %s",
					ruleno, replace(filter_template, "$client", *nets)
				);
			ret |= system(sb->buf);
		}
	} else {
		memset(&rule, 0, sizeof rule);
	
		rule.fw_number=ruleno;
		rule.fw_flg |= IP_FW_F_ACCEPT;
		rule.fw_prot = IPPROTO_TCP;
	
		/* 1 port of destination (3128) and all source ports */
		rule.fw_nports=0x10;
		rule.fw_uar.fw_pts[0]=3128;
	
   	     /* No direction specified -> do both directions */
		rule.fw_flg |= IP_FW_F_OUT | IP_FW_F_IN;
	
		for(; *nets; nets++) {
			if(fill_ip(&rule.fw_src, &rule.fw_smsk, *nets)) {
				ret |= 1;
			} else {
				ret |= setsockopt(sock, IPPROTO_IP,
					IP_FW_ADD, &rule, sizeof rule);
			}
		};
	};

	return ret;
};
#else
int
addrule(int ruleno, char **nets) {
	if(ruleno < 10000 || ruleno >= 20000 || !nets || !*nets) return 1;
	for(; *nets; nets++)
		printf("A %d allow %s to any 3128\n", ruleno, *nets);
	return 0;
};
#endif

#ifdef	HAVE_NETINET_IP_FW_H

int
redirector() {
	struct ip_fw rule;
	int ret = 0;
	char **nets;

	if(tail_template) {
		static sbuf *sb = NULL;

		if(!sb)
			sb = sbuf_init();
		else
			sbuf_zero(sb);

		sbuf_sprintf(sb, "/sbin/ipfw add 19999 %s", tail_template);
		ret |= system(sb->buf);
	} else {
		memset(&rule, 0, sizeof rule);
	
		rule.fw_number=19999;
		rule.fw_flg |= IP_FW_F_FWD;
		rule.fw_prot = IPPROTO_TCP;
	
		/* 1 port of destination (3128) and all source ports */
		rule.fw_nports=0x10;
		rule.fw_uar.fw_pts[0]=3128;
	
		rule.fw_fwd_ip.sin_len = sizeof(struct sockaddr_in);
		rule.fw_fwd_ip.sin_family = AF_INET;
		rule.fw_fwd_ip.sin_port = atoi(fwdtoport);
	
		rule.fw_tcpf |= IP_FW_TCPF_SYN;
		rule.fw_tcpnf |= IP_FW_TCPF_ACK;
	
		fill_ip(&(rule.fw_fwd_ip.sin_addr), NULL, fwdtoip);
	
   	     /* No direction specified -> do both directions */
		rule.fw_flg |= IP_FW_F_OUT | IP_FW_F_IN;
	
		for(nets = controlnets; *nets; nets++) {
			if(fill_ip(&rule.fw_src, &rule.fw_smsk, *nets))
				ret |= 1;
			else {
				ret |= setsockopt(sock, IPPROTO_IP,
				IP_FW_ADD, &rule, sizeof rule);
			};
		};
	}
	
	return ret;
};

#endif
