#include <err.h>
#include <strfunc.h>
#include <lgba.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/param.h>

char *atr[]={
	"createdby",
	"createtime",
	"modifiedby",
	"modifytime",
	NULL};

#define BACKUP_PATH	"/usr/swap/backup/users/"
#define MAIL_PATH	"/var/mail/"

#define CM_RM 		"/bin/rm"
#define CM_TAR 		"/usr/bin/tar"

int m_validattr(char *p);
int crfunc(int);

int save_attr(char *login, char *dn);
int save_home(char *p);
int kill_object(char *p);

int main(int ac, char *av[])
{
	char *at;

	if(ac != 2)
		errx(0, "usage: ldmu user\n");

	if(lgba_init(NULL))
		errx(1, "LGBA Init error.");

	if(lgba_open())
		errx(1, "LGBA Open error.");

	lgba_setuser("root");

	atr[0]=getmodifiedattributename();
	atr[1]=getmodifierattributename();
	atr[2]=getcreatorattributename();
	atr[3]=getcreatedattributename();

	at=sf_malloc(strlen(lgba_ObjectAttribute()) + strlen(av[1]) + 4);
	sprintf(at, "(%s=%s)", lgba_ObjectAttribute(), av[1]);

	lgba_search(NULL, at, NULL, NULL, NULL,
		NULL, crfunc, 0);

	free(at);

	return 0;
}

int m_validattr(char *p)
{
	char **q = atr;
	for( ; *q ; q++ ) if(!strcasecmp(*q, p)) return 0;
	return 1;
}

char *k_path(char *qu, char *q)
{
	if(!q || !*q){ *qu=0; return ""; }
	if(!realpath(q,qu)){ *qu=0; return ""; }

	if(qu[strlen(qu)-1]=='/')
		qu[strlen(qu)-1]=0;

	return qu;
}

char *get_path(char *q)
{
	static char buff[MAXPATHLEN];
	char *r;

	k_path(buff,q);

	r = strrchr(buff,'/');
	if( r!=NULL ) *r = 0;

	return buff;
}

char *get_file(char *q)
{
	static char buff[MAXPATHLEN];
	char *r;

	k_path(buff,q);

	r = strrchr(buff,'/');
	if( r!=NULL ) return r+1;

	return buff;
}

int crfunc(int zzz) {
	char *objectname=lgba_objectname();
	char *objectdn=lgba_objectdn();

	printf("%s:\n", objectdn);

	if(save_attr(objectname, objectdn)
	  && save_home(objectname)
	  && kill_object(objectname) )
		printf("-=> OK\n");
	else
		printf("-=> FAILED\n");

	return 0;
};

int save_attr(char *login, char *dn)
{
	char **al;
	char **p;
	char pt[4096];
	FILE *F;

	if( ! (al = lgba_objectattributes()) ) {
		warnx("ERROR: cannot get object attributes.");
		return 0;
	}

	strcpy(pt,BACKUP_PATH);
	strcat(pt, login);
	strcat(pt, ".nfo");

	if(!(F=fopen(pt, "w"))) {
		warnx("ERROR: cannot open file: %s", pt);
		return 0;
	};

	fprintf(F, "dn: %s\n",dn);

	for( p = al; *p ; p++ ) {
		register char **vp;
		register slist *sl;

		if( !m_validattr(*p) ) continue;

		sl=getattr2(*p);
		if(!sl)
			continue;
		for(vp=sl->list; *vp; vp++)
			fprintf(F,"%s: %s\n", *p, *vp);

	};

	fclose(F);

	return 1;
}

int save_home(char *p) {
	struct stat t;
	char pt[4096];
	char *home=getattr("home");
	int rv;

	if(stat(home,&t) || !*home ||
		!strcmp(home,"/") ||
		!strncmp(home,"/no", 3) ||
		!strncmp(home,"/var",4) ||
		!strncmp(home,"/dev",4)
	) {
		warnx("HOME[%s] not saved\n", home);
		return 0;
	};

	//...
	sprintf(pt,"cd %s; %s -cyvpf %s%s_home.tar.bz2 %s",
			get_path(home),CM_TAR,BACKUP_PATH,p,get_file(home));
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	sprintf(pt,"%s -rf %s",CM_RM,home);
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	printf("user mail:\n");
	sprintf(pt,"cd %s; %s -cyvpf %s%s_mail.tar.bz2 %s",
			MAIL_PATH,CM_TAR,BACKUP_PATH,p,p);
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	sprintf(pt,"/bin/rm -f %s%s %s.%s.pop",MAIL_PATH,p,MAIL_PATH,p);
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	
	return 1;
}

int kill_object(char *p)
{
	printf("removing object... ");
	if(rmobject(p)) {
		printf("FAILED\n");
		lgba_perror("ERROR");
		return 0;
	}
	printf("ok");
	return 1;
}


