#include <err.h>
#include <strfunc.h>
#include <unistd.h>
#include <lgba.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/param.h>

char *atr[]={
	"createdby",
	"createtime",
	"modifiedby",
	"modifytime",
	NULL
};

#define BACKUP_PATH	"/usr/swap/backup/users/"
#define MAIL_PATH	"/var/mail/"

#define CM_RM 		"/bin/rm"
#define CM_TAR 		"/usr/bin/tar"

int m_validattr(char *p);
int crfunc(int);
int existsfiles(char *p);

int restore_object(char *p);
int restore_files(char *p);

typedef struct sa {
	char *attr;
	char *value;
	struct sa *next;
}ta;

void kill_ta(ta *l);

int main(int ac, char *av[])
{
	char *at;

	if(ac != 2)
		errx(0, "usage: lrmu user");

	if(!existsfiles(av[1]))
		errx(1, "Error: user files not found\n");

	if(lgba_init(NULL))
		errx(1, "LGBA Init failed.");

	if(lgba_open())
		errx(1, "LGBA Open failed.");

	lgba_setuser("root");

	at=sf_malloc(strlen(lgba_ObjectAttribute()) + strlen(av[1]) + 4);
	sprintf(at, "(%s=%s)", lgba_ObjectAttribute(), av[1]);

	atr[0]=getmodifiedattributename();
	atr[1]=getmodifierattributename();
	atr[2]=getcreatorattributename();
	atr[3]=getcreatedattributename();

	lgba_search(NULL, at, NULL, NULL, NULL, NULL, crfunc, 0);

	if(restore_object(av[1]) && restore_files(av[1]))
		printf("OK\n");
	else
		printf("FAILED\n");

	return 0;
}

int
m_validattr(char *p) {
	char **q;

	for(q = atr; *q; q++)
		if(!strcasecmp(*q, p))
			return 0;

	return 1;
}

char *k_path(char *qu, char *q) {
	if(!q || !*q){ *qu=0; return ""; }
	if(!realpath(q,qu)){ *qu=0; return ""; }

	if(qu[strlen(qu)-1]=='/')
		qu[strlen(qu)-1]=0;

	return qu;
}

char *get_path(char *q) {
	static char buff[MAXPATHLEN];
	char *r;

	k_path(buff,q);

	r = strrchr(buff,'/');
	if( r!=NULL ) *r = 0;

	return buff;
}

char *get_file(char *q) {
	static char buff[MAXPATHLEN];
	char *r;

	k_path(buff,q);

	r = strrchr(buff,'/');
	if( r!=NULL ) return r+1;

	return buff;
}

int crfunc(int zzz) {
	errx(1, "ERROR: Object already exists.");
};

int existsfiles(char *u) {
	struct stat t;
	char pt[4096];

	if(countchar2(u,"-/.?~-!@#$%^&*()+|<>\\[]{}") != NULL)
		errx(0, "ERROR: illegal username.");

	strcpy(pt,BACKUP_PATH);
	strcat(pt,u);
	strcat(pt,".nfo");

	return !stat(pt,&t);
}

void kill_ta(ta *l)
{
	ta *p;
	while( l ){
		p = l->next;
		if(l->attr)free(l->attr);
		if(l->value)free(l->value);
		l = p;
	}
}

int restore_object(char *u)
{
	char dn[4096];
	char pt[4096];
	char ptv[4096];
	FILE *F;
	char *nn, *nv, *bdn;
	ta *ll=0, *lp;
	char *pc=0;
	char *lc=0, *p;

	strcpy(pt,BACKUP_PATH); strcat(pt,u); strcat(pt,".nfo");

	if(!(F=fopen(pt,"r"))){
		printf("ERROR: cannot open file: %s\n",pt);
		return 0;
	}

	if(fgets(dn,sizeof(dn)-1,F)==NULL){
		perror("ERROR(get dn)");
		fclose(F); return 0;
	}

	strtok(dn,"\r\n");
	printf("DN: %s\n",dn);
	nn=strtok(dn,"=");
	nv=strtok(NULL,",");
	bdn=strtok(NULL,""); while(bdn && *bdn==' ')bdn++;
	printf("node: (%s,%s)\n",nn,nv);
	printf("basedn:   \t(%s)\n",bdn);
	printf("currentdn:\t(%s)\n",lgba_current_dn());
	if(bdn && strcasecmp(bdn,lgba_current_dn())){
		printf("ERROR: wrong users' basedn\n");
		return 0;
	}

	// change current dn

	while(!feof(F)){
		if(fgets(pt,sizeof(pt)-1,F)==NULL) break;
		if(fgets(ptv,sizeof(ptv)-1,F)==NULL) break;
		strtok(pt,"\n\r"); strtok(ptv,"\n\r");
		if(!strcasecmp(pt,"objectclass")){
			lc = sf_strdup(ptv);
			if(strtok(ptv, ", ")!=NULL)
				pc = sf_strdup(ptv);
			continue;
		}
		//printf("%-20s = %-s\n",pt,ptv);
		if( !ll ){
			ll = lp = (ta *)sf_malloc(sizeof(ta));
		}
		else {
			lp->next = (ta *)sf_malloc(sizeof(ta));
			lp = lp->next;
		}
		bzero(lp,sizeof(ta));
		lp->attr = sf_strdup(pt);
		lp->value = sf_strdup(ptv);
	}
	fclose(F);

	printf("primary class: %s\n",pc);
	
	printf("creating object...\n");
 	if(create(nv,pc)){
		lgba_perror("ERROR");
		return 0;
	}

	if(!lgba_objectdn()){
		printf("ERROR: object not loaded :(\n");
		return 0;
	}
	printf("objectdn: %s\n",lgba_objectdn());

	printf("add classes:"); fflush(stdout);
	if( lc ) {
		p = strtok(lc,", ");
		for(	p = strtok(NULL,", ");
			p ;
			p = strtok(NULL, ", ")
		){
			printf("%s ",p);
			if(addclass( p )){
				printf("FAIL\n");
				lgba_perror("ERROR");
				return 0;
			}
		}
	}
	else printf("none.\n");

	printf("assigning attributes:\n");
	for( lp = ll; lp; lp=lp->next ) {
		printf("%s ",lp->attr);fflush(stdout);
		setattr(lp->attr,lp->value);
	}

	kill_ta(ll);
	if( pc ) free(pc);
	if( lc ) free(lc);

	if(lgba_save(1) || load(nv)) {
		lgba_perror("ERROR");
		return 0;
	};

	strcpy(pt,BACKUP_PATH); strcat(pt,u); strcat(pt,".nfo");
	unlink(pt);

	printf("object created successfuly\n");
	return 1;
}

int restore_files(char *p) {
	struct stat t;
	char pt[4096];
	char *home=getattr("home");
	int rv;

	if(!stat(home,&t) || !*home ||
		!strcmp(home,"/") ||
		!strncmp(home,"/no", 3) ||
		!strncmp(home,"/var",4) ||
		!strncmp(home,"/dev",4)
	) {
		printf("HOME[%s] not restored\n",home);
		return 0;
	}

	sprintf(pt,"%s%s_home.tar.bz2", BACKUP_PATH,p);

	if(stat(pt,&t)){
		printf("ERROR: home files missing\n");
		return 0;
	}

	// 1. restore home
	printf("restore users' home:\n");
	sprintf(pt,"cd %s; %s -xyvpf %s%s_home.tar.bz2 %s",
			get_path(home),CM_TAR,BACKUP_PATH,p,get_file(home));
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	sprintf(pt,"%s -f %s%s_home.tar.bz2",CM_RM,BACKUP_PATH,p);
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	// 2. restore mail
	printf("restore users' mail:\n");
	sprintf(pt,"cd %s; %s -xyvpf %s%s_mail.tar.bz2 %s",
			MAIL_PATH,CM_TAR,BACKUP_PATH,p,p);
	printf("[exec %s\n]",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);
	//
	sprintf(pt,"/bin/rm -f %s%s_mail.tar.bz2",BACKUP_PATH,p);
	printf("[exec %s]\n",pt);
	rv=system(pt);
	printf("[STATUS: %d]\n",rv);

	return 1;
};


