
#ifndef	__STDC__
#define	__STDC__
#endif

#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <lber.h>
#include <ldap.h>
#include <string.h>
#include <time.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef	HAVE_FNMATCH_H
#include <fnmatch.h>
#endif

#include <strfunc.h>

#include "lgba.h"
#include "friendly.h"
#include "readcf.h"
#include "pwd.h"
#include "file.h"
#include "getlists.h"
#include "intern.h"

extern answer loaded;
extern int lgba_errno;

extern LDAP *ld;

int
create(const char *obj, const char *cls) {
	acl *ac;
	char **sl;
	char **dl;
	int sms=lgba_i_smartsave;
	int incval;
	slist *clss=NULL;
	int os;

	/* Save modified data */
	lgba_save(1);

	ldap_flush_cache(ld);


	/* Simple checks */
	if(!cls) cls=lgba_i_node->defaultclass;
	if(!cls || !*cls) return lgba_errno=LGBA_VALUE;

	clss=sinit();

	sadd(clss, (char *)cls);

	if(obj) {
		if(!*obj) {
			if(lgba_i_node->inc_attr && ecq2(lgba_i_node->inc_attr, lgba_i_node->OA))
				obj=NULL;
			else
				return lgba_errno=LGBA_VALUE;
		};
	} else {
	  if(!lgba_i_node->inc_attr && !ecq2(lgba_i_node->inc_attr, lgba_i_node->OA))
		return lgba_errno=LGBA_VALUE;
	  obj=NULL;
	};

	if(obj && !lgba_i_validate(lgba_i_node->OA, obj))
		return lgba_errno=LGBA_VALUE;

	if(!(ac=lgba_i_getcreatelist(cls)))
		return lgba_errno=LGBA_PERM;

	sl=sarray(ac->attrlist, 0);
	dl=sarray(ac->defaults, 0);
	if((cfind(sl, "objectclass") == -1) || (cfind(sl, lgba_i_node->OA) == -1))
		return lgba_errno=LGBA_PERM;

	if(lgba_exists(obj) == 0) return lgba_errno=LGBA_ALEX;

	loaded.dn=lgba_internal_gendn(obj);

	os=0;
	lgba_i_smartsave=1;
	while(*sl) {
		if(ecq2(*sl, lgba_i_node->OA)) {
			if(os) continue; else os++;

			if(!obj) {
				if((incval=lgba_i_getincrementedvalue()) == -1) {
					lgba_free();
					return lgba_errno=LGBA_PERM;
				};
				if(loaded.dn) free(loaded.dn);
				loaded.dn=lgba_internal_gendn(genstr(incval));
				obj=sf_strdup(genstr(incval));
			};

			lgba_i_addmodchain(LDAP_MOD_ADD,
			sf_strdup(*sl), charize(obj));

			sl++; dl++;
			continue;
		} else if(ecq2(*sl, "objectclass")) {
			if(lgba_i_getaccess(*dl, "objectclass", AC_CREATE) == 1) {
				/* allow */
				if(scfind(clss, *dl) == -1)
					sadd(clss, *dl);
			};
			sl++; dl++; continue;
		};
		lgba_i_addmodchain(LDAP_MOD_ADD, sf_strdup(*sl), charize(*dl));
		sl++; dl++;
	};

	lgba_i_addmodchain(LDAP_MOD_ADD, sf_strdup("objectclass"), mkarray(clss, 0));

	if(lgba_i_node->mnt_creator)
		lgba_i_addmodchain(LDAP_MOD_ADD, sf_strdup(lgba_i_node->mnt_creator),
			charize(lgba_i_user));
	if(lgba_i_node->mnt_created)
		lgba_i_addmodchain(LDAP_MOD_ADD, sf_strdup(lgba_i_node->mnt_created),
			charize(timetostr(time(NULL), TFMT_X208)));

	if(lgba_i_node->inc_attr && !ecq2(lgba_i_node->inc_attr, lgba_i_node->OA)) {
		if((incval=lgba_i_getincrementedvalue()) == -1) {
			lgba_free();
			return lgba_errno=LGBA_PERM;
		};
		lgba_i_addmodchain(LDAP_MOD_ADD,
			sf_strdup(lgba_i_node->inc_attr),
			charize(genstr(incval)));
	};

	lgba_i_smartsave=sms;

	loaded.objectclass=clss;
	loaded.object=sf_strdup(obj);

	if(lgba_save(1))
		return lgba_errno;

	if((os=lgba_exists(obj)) == LGBA_TOOMANY) {
		char *dn=lgba_internal_gendn(obj);
		ldap_delete_s(ld, dn);
		free(dn);
		return lgba_errno=LGBA_ALEX;
	};

	return load(obj);
};

