#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "lgba.h"
#include "readcf.h"
#include "intern.h"
#include <strfunc.h>

int makefd=1;
char *fmt;

char *lgba_i_file_before=NULL;
char *lgba_i_file_after=NULL;

int *connectchild(const char *cmd);

int mkf_display(int z) { format(fmt, makefd); return 0; };

int
makefile(file *fl) {
	char *tmpfile;
	char *p;

	if(!fl) return LGBA_VALUE;

	printf("File: %s [%s]\n", fl->name, fl->filename);

	if(fl->node && lgba_i_node && strcmp(fl->node, lgba_i_node->name)) {
		if(lgba_chnode(fl->node)) {
			printf("\tERROR changing node to '%s'\n", fl->node);
			return LGBA_VALUE;
		};
	};
	if(!fl->node || !lgba_i_node) {
		printf("\tERROR changing node\n");
		if(lgba_chnode(lgba_i_defaultnode)) {
			printf("\tERROR changing node to '%s'\n", lgba_i_defaultnode);
		};
		return LGBA_VALUE;
	};

	/* Strip space */
	for(p=fl->filename; isspace(*p); p++);

	/* Is this a process? */
	if(*p == '|') {
		int *fd, st;
		if(!(fd=connectchild(++p))) {
			_lgba_internal_warn("Can't open pipes.");
			return LGBA_VALUE;
		};
		makefd=fd[0];
		format(fl->head, makefd);
		fmt=fl->body;
		lgba_search(NULL, fl->filter, fl->attrlist, fl->sortby, NULL, NULL, mkf_display, 0);
		format(fl->tail, makefd);
		close(fd[0]);
		close(fd[1]);
		waitpid(fd[2], &st, 0);
		makefd=1;
		if(fl->program)
			xsystem(fl->program);
		return 0;
	};

	if(getuid() && fl->owner != getuid()) {
		printf("\tERROR: insufficient rights\n");
		return LGBA_PERM;
	};

	tmpfile=(char *)sf_malloc(strlen(fl->filename) + 16);
	sprintf(tmpfile, "%s.lgba.%ld", fl->filename, (long)getpid());

	if((makefd=open(tmpfile, O_CREAT | O_WRONLY | O_TRUNC, 0)) == -1) {
		printf("\tERROR: Can't open temporary file.\n");
		free(tmpfile);
		return LGBA_COMMAND;
	}

	if(fchmod(makefd, fl->mode) == -1) {
		printf("\tERROR: can't fchmod() file.\n");
		close(makefd); unlink(tmpfile); free(tmpfile);
		return LGBA_PERM;
	}

	if(fchown(makefd, fl->owner, fl->group) == -1) {
		printf("\tERROR: can't fchown() file.\n");
		close(makefd); unlink(tmpfile); free(tmpfile);
		return LGBA_PERM;
	}

	if(fl->append) {
		int fdfrom;
		ssize_t size;
		if((fdfrom=open(fl->filename, O_RDONLY)) == -1) {
			char buf[16384];
			while((size=read(fdfrom, buf, sizeof(buf))))
				write(makefd, buf, size);
			close(fdfrom);
		}
	}

	format(fl->head, makefd);
	fmt=fl->body;
	lgba_search(NULL, fl->filter, fl->attrlist, fl->sortby, NULL, NULL, mkf_display, 0);
	format(fl->tail, makefd);

	fsync(makefd);
	close(makefd);

	if(rename(tmpfile, fl->filename)) {
		unlink(tmpfile); free(tmpfile);
		printf("\tERROR: rename() failed.\n");
		return LGBA_PERM;
	};

	if(fl->program)
		xsystem(fl->program);

	return 0;
};

int makefiles(char **vp) {
	file *fl=lgba_i_file_head;
	char *nn=lgba_i_node->name;
	int ntr=0;

	if(!fl) return LGBA_UNKNOWN;

	if(!vp || !*vp) {
		ntr=1;
		if(lgba_i_files) vp=sarray(lgba_i_files, 0);
	};

	if(ntr) xsystem(lgba_i_file_before);

	while(fl) {
		if(!vp || !*vp || cfind(vp, fl->name) != -1) makefile(fl);
		fl=fl->next;
	};

	if(ntr) xsystem(lgba_i_file_after);

	/* back to USSR */
	if(!lgba_i_node || strcmp(lgba_i_node->name, nn))
		lgba_chnode(lgba_i_node->name);

	return 0;
};

/* returns pipe[write, read] */
int *
connectchild(const char *cmd) {
	static int pipes[3];
	int mother_wr[2];
	int mother_rd[2];
	int childpid;

	if(pipe(mother_wr) || pipe(mother_rd))
		{ printf("ERROR: Can't open pipe().\n"); return NULL; };

	fflush(NULL);

	if((childpid=fork())==-1) {
		printf("ERROR: Failed due to fork() failure.\n");
		return NULL;
	};

	/* child */
	if(!childpid) {
		close(mother_wr[1]);
		close(mother_rd[0]);
		close(0);
		dup(mother_wr[0]);
		close(1);
		dup(mother_rd[1]);
		system(cmd);
		exit(2);
	};

	/* mother */
	close(mother_wr[0]);
	close(mother_rd[1]);

	*(pipes)=mother_wr[1];
	*(pipes+1)=mother_rd[0];
	*(pipes+2)=childpid;

	/* pipes[write, read] */
	return pipes;
};

int
xsystem(const char *cmd) {
	int pid, st;
	int rc;
	int nfd;

	if(!cmd)
		return 1;

	fflush(NULL);

	nfd = open("/dev/null", O_RDONLY);
	if(nfd == -1) {
		printf("ERROR: Failed due to failure to open /dev/null.\n");
		return 1;
	}

	if((pid=fork())==-1) {
		printf("ERROR: Failed due to fork() failure.\n");
		return 1;
	};

	if(!pid) {      // child
		close(2);	// close stderr
		dup(1);		// reconnect stdout
		close(0);	// close stdin
		dup(nfd);	// open stdin
		rc = system(cmd);
		printf("['%s' completed with code %d]\n", cmd, rc);
		exit(2);
	} else {	// parent
		close(nfd);
	};

	do {
		if(waitpid(pid, &st, 0) == -1)
			break;
	} while(!WIFSTOPPED(st));

	return 0;
};


