#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <lber.h>
#include <ldap.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include <strfunc.h>

#include "readcf.h"
#include "friendly.h"
#include "lgba.h"
#include "intern.h"

#define DEFLISTFORMAT ", "
char *listformat=DEFLISTFORMAT;

int format(char *fmt, int fd) {
	fmt=getformat(fmt);
	if(!fmt) return -1;
	return write(fd, fmt, strlen(fmt));
}

char **lgba_internal_friendly_wrap(char *s, void *key) {
	static char *z[]={ NULL, NULL };
	(void)key;
	*z=lgba_map_friendly(s);
	return z;
};

char *getformat(char *fmt) {
	static fmt_base *fb=NULL;

	if(!fb) {
		fb=format_init();
		format_metarule(fb, '{', '}', lgba_internal_override);
		format_metarule(fb, '`', '\'', lgba_internal_friendly_wrap);
	};

	return formatf(fb, fmt, NULL);
};

char *
getattr(const char *attr) {
	register slist *sl=getattr2(attr);

	if(!sl || !sl->count) return "";

	if(sl->count == 1)
		return sl->list[0];

	if(countchar2(listformat, "$\"\\"))
		return sjoin(sl, getformat(listformat));

	return sjoin(sl, listformat);
};

slist *
getattr2_s(answer *answer, const char *attr) {
	register avt **avp;
	register avt *ap;
	register char *z, *s;
	register int i;
	register int ty;

	if(!attr || !answer) return NULL;
	if(!answer->list) return NULL;

	avp = answer->list;

	ty = lgba_gettype(attr);

	for(; (ap = *avp); avp++) {
		if(ecq(ap->key, attr)) {
			if(ty & TY_BINARY)
				return ap->vals;
			if(lgba_i_htmlmode && !ap->dequoted) {
				if(!ap->vals->count)
					return ap->vals;
				for(i=0; i < ap->vals->count; i++) {
					z=html_quote(s=ap->vals->list[i]);
					if(z == s)
						continue;
					free(s);
					ap->vals->list[i]=z;
					ap->vals->lens[i]=strlen(z);
					html_quote_detach();
				};
				ap->dequoted=1;
			};
			return ap->vals;
		};
	};
	return NULL;
};


slist *
getattr2(const char *attr) {
	register avt **avp=loaded.list;
	register avt *ap;
	register char *z, *s;
	register int i;
	register int ty;

	if(!attr) return NULL;
	if(!avp) return NULL;

	ty = lgba_gettype(attr);

	for(; (ap = *avp); avp++) {
		if(ecq(ap->key, attr)) {
			if(ty & TY_BINARY)
				return ap->vals;
			if(lgba_i_htmlmode && !ap->dequoted) {
				if(!ap->vals->count)
					return ap->vals;
				for(i=0; i < ap->vals->count; i++) {
					z=html_quote(s=ap->vals->list[i]);
					if(z == s)
						continue;
					free(s);
					ap->vals->list[i]=z;
					ap->vals->lens[i]=strlen(z);
					html_quote_detach();
				};
				ap->dequoted=1;
			};
			return ap->vals;
		};
	};
	return NULL;
};

char **
lgba_i_httpize(char **vp, char *at) {
	static slist *sl=NULL;

	if(!vp || !lgba_i_htmlmode) return vp;
	if(!sl) sl=sinit();

	sclear(sl);
	while(*vp) {
		splitf(sl, *vp, ", ", 0);
		vp++;
	}
	if(!sl->count) return NULL;

	if(!lgba_i_htmlmode) return sarray(sl, 0);

	vp=sl->list;
	while(*vp) {
		sbuf *sb;

		sb = sbuf_init();

		/* URL or mailto? */
		if(strchr(*vp, '/'))
		    sbuf_sprintf(sb, "<A HREF=\"%s\">%s</A>",	
			*vp, *vp);
		else
		    sbuf_sprintf(sb, "<A HREF=\"mailto:%s\">%s</A>",
			*vp, *vp);

		free(*vp);
		*vp=sb->buf;
		free(sb);

		vp++;
	}

	return sarray(sl, 0);
}

char **
lgba_i_nbspize(char **vp, char *at) {
	static slist *sl=NULL;
	static slist *tl=NULL;

	if(!vp || !lgba_i_htmlmode) return vp;

	if(!sl) {
		sl=sinit();
		tl=sinit();
	} else sclear(sl);

	while(*vp) {
		sclear(tl);
		splitf(tl, *vp, " \t", 0);
		vp++;
		sadd(sl, sjoin(tl, "&nbsp;"));
	}

	return sarray(sl, 0);
};

char **lgba_i_timeize(char **in, char *at) {
	static slist *sl=NULL;

	if(!in || !*in) return in;

	if(!sl) sl=sinit();
	else sclear(sl);

	for(; *in; in++) {
		sadd(sl,
			timetostr(strtotime(*in),
				TFMT_LOCAL | TFMT_UF | TFMT_RFC822)
		);
	};

	if(!sl->count) return in;

	return sarray(sl, 0);
};


char **lgba_i_mapize(char **in, char *at) {
	static slist *sl=NULL;
	slist *a=NULL;
	slist *v=NULL;
	char *s;

	if(!in || !*in) return in;

	s=lgba_gettypespec(at);
	if(!s) return in;

	if(!sl) sl=sinit();
	else sclear(sl);

	if(lgba_readavfile(s, &a, &v)) return in;

	for(; *in; in++) {
		s=sget2(a, *in, v);
		if(!s) sadd(sl, *in);
		else sadd(sl, s);
	};

	sfree(a);
	sfree(v);

	return sarray(sl, 0);
};


