#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>

#include <ctype.h>

#include <lber.h>
#include <ldap.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include "lgba.h"
#include "friendly.h"
#include "readcf.h"
#include "pwd.h"
#include "file.h"
#include "getlists.h"
#include "intern.h"

answer loaded;
int lgba_i_counter;
int lgba_errno=0;

char *_lgba_i_loadfilter = NULL;

int lgba_i_addmodchain(int, char *, void *);
int lgba_cd_single(char *);
char *lgba_internal_gendn(const char *);

LDAP *ld;

char **lgba_internal_atl=NULL;

#if	( LDAP_API_VERSION > 2000 )

int
_lgba_internal_rebind(LDAP *ld,
		const char *url, int request, ber_int_t msgid) {
	if(!lgba_i_node) return LDAP_LOCAL_ERROR;

	return ldap_bind_s(ld, lgba_i_node->binddn, lgba_i_node->bindpwd, LDAP_AUTH_SIMPLE);
}

#else	/* OpenLDAP rebind extension */

int
_lgba_internal_rebind(LDAP *ld,
		char **whop, char **credp, int *methodp, int freeit) {
	if(freeit) return LDAP_SUCCESS;
	if(!lgba_i_node) return LDAP_LOCAL_ERROR;

	*whop=lgba_i_node->binddn;
	*credp=lgba_i_node->bindpwd;
	*methodp=LDAP_AUTH_SIMPLE;

	return LDAP_SUCCESS;
}

#endif	/* OpenLDAP rebind extension */


void lgba_setattrlist(char **atl) {
	if(lgba_internal_atl)
		free_values(lgba_internal_atl);

	if(!atl) { lgba_internal_atl=NULL; return; };

	copy_values(atl, &lgba_internal_atl);
	if(cfind(lgba_internal_atl, lgba_i_node->OA) == -1)
		lgba_internal_atl[count_values(lgba_internal_atl)]=sf_strdup(lgba_i_node->OA);
	if(cfind(lgba_internal_atl, "objectclass") == -1)
		lgba_internal_atl[count_values(lgba_internal_atl)]=sf_strdup("objectclass");
};

char *lgba_ObjectAttribute() { return lgba_i_node->OA; };

char *lgba_err2string(int err) {
	switch(err) {
		case 0: return "No error";
		case 1: return "Unknown error";
		case 2: return "Config error";
		case 3: return "LDAP connection error";
		case 4: return "Permission denied";
		case 5: return "LDAP error";
		case 6: return "Command syntax/order error";
		case 7: return "Set language failed";
		case 8: return "Access denied";
		case 9: return "Object not found";
		case 10: return "Modify failed";
		case 11: return "LDAP bind error";
		case 12: return "Input value error";
		case 13: return "No such class";
		case 14: return "Too many values";
		case 15: return "Create object failed";
		case 16: return "Node not found";
		case 17: return "Already defined";
		case 18: return "Already exists";
		case 19: return "Insufficient increase range";
		case 20: return "No such attribute";
		case 21: return "Too many found";
	}
	return "Unknown error";
}

int
lgba_init(char *config) {
	static int z=0;
	char *c;

	if(strfunc_version() < 722) {
		_lgba_internal_warn("Please upgrade your libstrfunc library to 7.2.2.\nRefer to http://www.spelio.net.ru/soft");
		return LGBA_UNKNOWN;
	};

	if(z)
		return lgba_errno=LGBA_ALREADY;

	if(lgba_readcf(config))
		return lgba_errno=LGBA_CONF;

	if((c = getlogin()))
		lgba_i_user = sf_strdup(c);
	else
		lgba_i_user = sf_strdup("nobody");

	if(!lgba_i_user)	/* Resource shortage */
		return LGBA_RSHORT;

	for(c = lgba_i_user; *c; c++)
		*c = tolower(*c);

	z=1;

	return LGBA_NOERR;
};

int
lgba_open() {
	char **vp=sarray(lgba_i_node->bindserver, 0);
	ld=NULL;
	if(!vp) return lgba_errno=LGBA_CONF;
	while(*vp)
		if((ld=ldap_open(*vp++, 389))) break;

	if(!ld) return lgba_errno=LGBA_CONN;

	/* Caching is broken in OpenLDAP 2.x */
#if ! defined ( LDAP_VERSION3 ) || ! defined ( LDAP_API_FEATURE_X_OPENLDAP )
	ldap_enable_cache(ld, 30, 1024*1024);
#else
#warning
#warning CACHING disabled due to OpenLDAP [censored].
#warning
#endif

	if(ldap_simple_bind_s(ld, lgba_i_node->binddn, lgba_i_node->bindpwd))
		lgba_perror("~WARN: LDAP bind failed: %s\n");

#if defined ( LDAP_VERSION3 ) && ! defined ( LDAP_API_FEATURE_X_OPENLDAP )
	ldap_set_rebind_proc(ld, _lgba_internal_rebind, NULL);
#else
	ldap_set_rebind_proc(ld, _lgba_internal_rebind);
#endif

	return LGBA_NOERR;
};

int
lgba_close() {
	int ret;

	if(!ld) return 0;

	ret=lgba_save(1);
	ldap_unbind(ld);
	ld=NULL;
	return ret;
};

void
lgba_free_s(answer *answer) {

	if(!answer) return;
	if(!answer->dn) return;

	if(answer->lm_list)
		lgba_internal_destroymodchain_s(answer, 0);

	/* We must not deal with pointers. */
	if(answer->protected) {
		memset(answer, 0, sizeof(*answer));
		return;
	};

	while(answer->count--) {
		free(answer->list[answer->count]->key);
		sfree(answer->list[answer->count]->vals);
		free(answer->list[answer->count]);
	};
	answer->count=0;

	if(answer->dn) { free(answer->dn); answer->dn=NULL; }
	if(answer->object) { free(answer->object); answer->object=NULL; }
	if(answer->objectclass) {
		sfree(answer->objectclass);
		answer->objectclass=NULL;
	};
	answer->lfa=NULL;
};

void
lgba_free() {
	lgba_free_s(&loaded);
};

int
lgba_save(int dofree) {

	if(!loaded.dn) return 0;

	if(!dofree)
		return lgba_internal_destroymodchain_s(&loaded, 1);

	if(loaded.lm_list)
		lgba_internal_destroymodchain_s(&loaded, 1);

	lgba_free_s(&loaded);

	return 0;
};

int
load(const char *object) {
	LDAPMessage *res, *en;
	char *filter;
	int k;

	if(!ld && lgba_open())
		return lgba_errno;
	if(!object)
		return lgba_errno=LGBA_VALUE;

	lgba_save(1); /* Save modified object prior to load another */

	filter=(char *)alloca(
		strlen(lgba_i_node->OA)
		+ strlen(object)
		+ (_lgba_i_loadfilter?strlen(_lgba_i_loadfilter):0)
		+ 8
	);
	if(!filter) {
		lgba_errno = LGBA_UNKNOWN;
		return -1;
	}
	if(_lgba_i_loadfilter) {
		if(*_lgba_i_loadfilter == '(') {
			sprintf(filter, "(&(%s=%s)%s)", lgba_i_node->OA, object,
				_lgba_i_loadfilter);
		} else {
			sprintf(filter, "(&(%s=%s)(%s))", lgba_i_node->OA, object,
				_lgba_i_loadfilter);
		}
	} else {
		sprintf(filter, "(%s=%s)", lgba_i_node->OA, object);
	}

	k=ldap_search_s(ld, lgba_current_dn(), LDAP_SCOPE_ONELEVEL, filter,
		lgba_internal_atl, 0, &res);

#ifdef	LDAP_NAME_ERROR
	if(LDAP_NAME_ERROR(k))
		return lgba_errno=LGBA_NOTFOUND;
#else
	if(NAME_ERROR(k))
		return lgba_errno=LGBA_NOTFOUND;
#endif

	switch(k) {
	case 0:
		break;
	case LDAP_SERVER_DOWN:
		if(lgba_open())
			return lgba_errno;
		return load(object);
	default:
		return lgba_errno=LGBA_LDAP;
	};

	/* How many found? */
	if((k=ldap_count_entries(ld, res)) != 1) {
		ldap_msgfree(res);
		if(k == 0) lgba_errno=LGBA_NOTFOUND;
		else if(k > 1) lgba_errno=LGBA_TOOMANY;
		else lgba_errno=LGBA_LDAP;
		return lgba_errno;
	}

	/* Filling an answer. */
	if(!(en=ldap_first_entry(ld, res))) {
		ldap_msgfree(res);
		return lgba_errno=LGBA_LDAP;
	};

	lgba_i_fillanswer(&loaded, en);

	ldap_msgfree(res);
	return LGBA_NOERR;
}

char *prepend_(char *s) {
	static sbuf *sb = NULL;

	if(!sb)
		sb = sbuf_init();
	else
		sbuf_zero(sb);

	sbuf_add(sb, "_");
	sbuf_add(sb, s);

	return sb->buf;
};

int
lgba_countwithattribute(char *a, char *v) {
	char *flt;
	LDAPMessage *res;
	char *atl[2]={ a, NULL };
	int n;

	if(!a || !v) return 0;

	flt=sf_malloc(strlen(a) + strlen(v) + 4);
	sprintf(flt, "%s=%s", a, v);

	if(ldap_search_s(ld, lgba_current_dn(), LDAP_SCOPE_ONELEVEL,
	  flt, atl, 1, &res)) return 0;
	n=ldap_count_entries(ld, res);
	ldap_msgfree(res);

	free(flt);
	return n;
};

int
lgba_i_getincrementedvalue() {
	LDAPMessage *res;
	int incval;
	char **vp;
	char *s=prepend_(lgba_i_node->inc_attr);
	char *atl[] = { s, NULL };
	LDAPMod *lm;
	LDAPMod *lma[]={ NULL, NULL };
	int ec;
	int loops=0;

	incval=lgba_i_node->inc_start;

	if(!ldap_search_s(ld, lgba_current_dn(), LDAP_SCOPE_BASE,
	  "(|(!(objectclass=*))(objectclass=*))", atl, 0, &res)) {
		vp=ldap_get_values(ld, res, s);
		if(vp) {
			if(*vp) incval=atoi(*vp)+1;
			ldap_value_free(vp);
		};
		ldap_msgfree(res);
	};

	while(lgba_countwithattribute(lgba_i_node->inc_attr, genstr(incval)))
		if(++incval > lgba_i_node->inc_end) {
			if(loops++) return -1;
			incval=lgba_i_node->inc_start;
		};

	lm=(LDAPMod *)sf_malloc(sizeof(LDAPMod));
	lm->mod_op=LDAP_MOD_REPLACE;
	lm->mod_type=s;
	lm->mod_values=(char **)charize(genstr(incval));
	lma[0]=lm;

	ec=ldap_modify_s(ld, lgba_current_dn(), lma);

	free(lm->mod_values);
	free(lm);
	return incval;
};

int
lgba_internal_setobjects(char **objs) {
	char *rdn;
	char **vp;
	slist *wo;
	char *on;
	int k;

	if(!objs || !*objs || !**objs) {
		free_values(objs);
		return lgba_errno=LGBA_VALUE;
	};
	if(!loaded.dn)
		return lgba_errno = LGBA_NOTFOUND;

	lgba_save(0);

	if(!(wo=getattr2(lgba_i_node->OA))) {
		lgba_errno=LGBA_UNKNOWN;
		lgba_perror("-ERR: Error getting an Object attribute");
		return lgba_errno;
	};

	vp=objs-1;
	while(*(++vp)) {
		if(strchr(*vp, ',')) {
			free_values(objs);
			return lgba_errno=LGBA_VALUE;
		};
		if(scfind(wo, *vp) != -1) continue;
		if(lgba_exists(*vp) == 0)
			return lgba_errno=LGBA_ALEX;
	};

	if(!ecq(loaded.object, *objs)) {
		rdn=(char *)sf_malloc(strlen(*objs) + strlen(lgba_i_node->OA) + 10);
		sprintf(rdn, "%s=%s", lgba_i_node->OA, *objs);

#if defined ( LDAP_VERSION3 ) && ! defined ( LDAP_API_FEATURE_X_OPENLDAP )
		if((k=ldap_rename_s(ld, loaded.dn, rdn, NULL, 1, NULL, NULL))) {
#else
		if((k=ldap_modrdn2_s(ld, loaded.dn, rdn, 1))) {
#endif

			free(rdn);
			free_values(objs);
			_lgba_internal_warn("-ERR LDAP error %d/%s",
				k, ldap_err2string(k));
			return lgba_errno=LGBA_LDAP;
		};

		free(rdn);
		free(loaded.dn);
		loaded.dn=lgba_internal_gendn(*objs);
	};

	on=sf_strdup(*objs);
	lgba_i_addmodchain(LDAP_MOD_REPLACE, sf_strdup(lgba_i_node->OA), objs);
	if(lgba_save(1)) {
		free(on);
		return lgba_errno;
	};

	k=load(on);
	free(on);
	return k;
};

int
lgba_i_addmodchain(int op, char *ty, void *ar) {
	LDAPMod *lm;

	if(!loaded.dn)
		return lgba_errno=LGBA_COMMAND;

	if(!loaded.lm_list)
		loaded.lm_list = sinit();

	lm=(LDAPMod *)sf_malloc(sizeof(LDAPMod));

	lm->mod_op=op;
	lm->mod_type=ty;
	lm->mod_values=(char **)ar;

	sadd_attach(loaded.lm_list, (char *)lm, sizeof(LDAPMod));

	if(!lgba_i_smartsave)
		lgba_internal_destroymodchain_s(&loaded, 1);

	return LGBA_NOERR;
};

int
lgba_internal_destroymodchain_s(answer *answer, int dosave) {
	LDAPMod **lma;
	int ec;

	if(!answer->lm_list)
		return LGBA_NOERR;

	if(dosave) {
	  if(lgba_i_node->mnt_modifier)
		lgba_i_addmodchain(LDAP_MOD_REPLACE,
			sf_strdup(lgba_i_node->mnt_modifier), charize(lgba_i_user));
	  if(lgba_i_node->mnt_modified)
		lgba_i_addmodchain(LDAP_MOD_REPLACE,
			sf_strdup(lgba_i_node->mnt_modified),
			charize(timetostr(time(NULL), TFMT_X208)));
	};

	lma = (LDAPMod **)answer->lm_list->list;
	if(!lma)
		return LGBA_NOERR;

	if(!dosave) {
		ldap_mods_free(lma, 1);
		answer->lm_list->list=NULL;
		sfree(answer->lm_list);
		answer->lm_list=NULL;
		return 0;
	};

	if(!ld && lgba_open()) {
		ldap_mods_free(lma, 1);
		answer->lm_list->list=NULL;
		sfree(answer->lm_list);
		answer->lm_list=NULL;
		return lgba_errno;
	};

	if(!(*lma)->mod_op)
		ec=ldap_add_s(ld, answer->dn, lma);
	else
		ec=ldap_modify_s(ld, answer->dn, lma);

	ldap_uncache_entry(ld, answer->dn);

	switch(ec) {
		case 0:
			ldap_mods_free(lma, 1);
			answer->lm_list->list=NULL;
			sfree(answer->lm_list);
			answer->lm_list=NULL;
			return 0;
		case LDAP_SERVER_DOWN:
			if(!lgba_open())
				return lgba_internal_destroymodchain_s(&loaded, dosave);
			ldap_mods_free(lma, 1);
			answer->lm_list->list=NULL;
			sfree(answer->lm_list);
			answer->lm_list=NULL;
			return lgba_errno;
		case LDAP_ALREADY_EXISTS:
			ldap_mods_free(lma, 1);
			answer->lm_list->list=NULL;
			sfree(answer->lm_list);
			answer->lm_list=NULL;
			return lgba_errno=LGBA_ALEX;
		case LDAP_NO_SUCH_ATTRIBUTE:	/* No such attribute */
			ldap_mods_free(lma, 1);
			answer->lm_list->list=NULL;
			sfree(answer->lm_list);
			answer->lm_list=NULL;
			return lgba_errno=LGBA_NOATTR;
		case LDAP_INSUFFICIENT_ACCESS:
			ldap_mods_free(lma, 1);
			answer->lm_list->list=NULL;
			sfree(answer->lm_list);
			answer->lm_list=NULL;
			return lgba_errno=LGBA_ACCESS;
		default:
			_lgba_internal_warn("-ERR: LDAP: %d/%s", ec, ldap_err2string(ec));
	};

	ldap_mods_free(lma, 1);
	answer->lm_list->list=NULL;
	sfree(answer->lm_list);
	answer->lm_list=NULL;

	return lgba_errno=LGBA_LDAP;
};

char const *
lgba_current_node() {
	if(!lgba_i_node)
		return NULL;

	return lgba_i_node->name;
};

int
lgba_chnode(const char *name) {
	leafnode *ln=lgba_i_node_head;
	int k;

	lgba_dn_dirty=1;

	if(!name) name=lgba_i_defaultnode;

	while(ln) {
		if(ecq(ln->name, name)) break;
		ln=ln->next;
	};
	if(!ln) return lgba_errno=LGBA_NODENF;

	if(!ld ||
	!ecq(ln->bindserver->list[0], lgba_i_node->bindserver->list[0])) {
		LDAP *tld=NULL;
		char **vp=sarray(lgba_i_node->bindserver, 0);
		if(!vp) return lgba_errno=LGBA_CONF;
		while(*vp) if((tld=ldap_open(*vp++, 389))) break;
		if(!tld) return lgba_errno=LGBA_CONN;
		ldap_unbind(ld);
		ld=tld;

#if defined ( LDAP_VERSION3 ) && ! defined ( LDAP_API_FEATURE_X_OPENLDAP )
		ldap_set_rebind_proc(ld, _lgba_internal_rebind, NULL);
#else
		ldap_set_rebind_proc(ld, _lgba_internal_rebind);
#endif

	}

	lgba_i_node=ln;

	if((k=ldap_simple_bind_s(ld, lgba_i_node->binddn, lgba_i_node->bindpwd))) {
		if(k == LDAP_SERVER_DOWN) {
			if(lgba_open()) return lgba_errno;
			return lgba_chnode(name);
		}
		return lgba_errno=LGBA_BIND;
	}

	return LGBA_NOERR;
};

char *lgba_first_attribute() {
	if(!loaded.count) return NULL;
	if(!(loaded.lfa=loaded.list)) return NULL;
	loaded.lfa--;
	return lgba_next_attribute();
}

char *lgba_next_attribute() {
	if(!loaded.lfa)
		return lgba_first_attribute();

	loaded.lfa++;

	if(*loaded.lfa)
		return (*loaded.lfa)->key;

	loaded.lfa=NULL;

	return NULL;
}

int uploadattr(const char *attr, const char *val) {
	struct stat sb;
	size_t size;
	int fd;
	struct berval **bva;
	struct berval *bv;

	/* try as a file handle */
	if((fd=atoi(val)) <= 0)
		if((fd=open(val, O_RDONLY)) == -1) return lgba_errno=LGBA_VALUE;

	fstat(fd, &sb);
	size=sb.st_size;

	bva=(struct berval **)sf_malloc(sizeof(struct berval *) * 2);
	*bva=bv=(struct berval *)sf_malloc(sizeof(struct berval));
	bv->bv_val=(char *)sf_malloc(size);
	bv->bv_len=read(fd, bv->bv_val, size);
	bva[1]=NULL;

	return lgba_i_addmodchain(LDAP_MOD_REPLACE | LDAP_MOD_BVALUES, sf_strdup(attr), bva);
}

int
setattr(const char *attr, const char *val) {
	int tp;

	if(!attr || !*attr) return lgba_errno=LGBA_VALUE;
	if(ecq(attr, "objectclass")) return lgba_errno=LGBA_VALUE;
	if(!canIwrite(attr)) return lgba_errno=LGBA_PERM;
	if(!val) return rmattr(attr, NULL);
	if(!lgba_i_validate(attr, val)) return lgba_errno=LGBA_VALUE;

	if(ecq(attr, lgba_i_node->OA))
		return lgba_internal_setobjects(charize(val));

	tp=lgba_gettype(attr);
	if(tp & TY_NUMERIC)
		(char *)val=lgba_internal_numericvalue(attr, val);
	if(tp & TY_BINARY) return uploadattr(attr, val);
	if(tp & TY_PASSWD) (char *)val=lgba_i_makepassword(val, tp & TY_CRYPTPW);

	if(!val) return lgba_errno=LGBA_VALUE;
	return lgba_i_addmodchain(LDAP_MOD_REPLACE, sf_strdup(attr), charize(val));
};

int
addattr(const char *attr, char *val) {
	slist *sl;
	if(!attr || !*attr) return lgba_errno=LGBA_VALUE;
	sl=sinit();
	simport(sl, mkarray(getattr2(attr), 0));
	if(!sl->count) {
		sfree(sl);
		return lgba_errno=setattr(attr, val);
	};
	sadd(sl, val);
	lgba_errno=setattr2(attr, mkarray(sl, 0));
	sfree(sl);
	return lgba_errno;
};

int
addattr2(const char *attr, char **vals) {
	slist *sl;
	if(!attr || !*attr || !vals || !*vals) return lgba_errno=LGBA_VALUE;

	sl=sinit();
	simport(sl, mkarray(getattr2(attr), 0));
	if(!sl->count) {
		sfree(sl);
		return lgba_errno=setattr2(attr, vals);
	};
	simport(sl, vals);
	lgba_errno=setattr2(attr, mkarray(sl, 0));
	sfree(sl);
	return lgba_errno;
};

/* setattr2() frees **vals!!! */
int
setattr2(const char *attr, char **vals) {
	int tp;
	char *t;

	if(!attr || !*attr || ecq(attr, "objectclass")) {
		free_values(vals);
		return lgba_errno=LGBA_VALUE;
	};

	if(!vals) {
		free_values(vals);
		return rmattr(attr, NULL);
	};

	if(!*vals) {
		free_values(vals);
		vals=charize("");
	};

	if(!canIwrite(attr)) { free_values(vals); return lgba_errno=LGBA_PERM; }

	tp=lgba_gettype(attr);
	if((tp & TY_SINGLE) && vals && *vals && *(vals+1)) {
		free_values(vals);
		return lgba_errno=LGBA_TOOMANY;
	};
	if(!lgba_i_validate2(attr, (const char **)vals)) {
		free_values(vals);
		return lgba_errno=LGBA_VALUE;
	};

	if(tp & TY_BINARY) return uploadattr(attr, *vals);

	if(ecq(attr, lgba_i_node->OA))
		return lgba_internal_setobjects(vals);

	if (tp & TY_NUMERIC) {
		t=lgba_internal_numericvalue(attr, *vals);
		free(*vals);
		*vals=sf_strdup(t);
	};

	if(tp & TY_PASSWD) {
		vals[1]=sf_strdup(lgba_i_makepassword(*vals, tp & TY_CRYPTPW));
		if(!vals[1]) {
			free_values(vals);
			return lgba_errno=LGBA_VALUE;
		};
		free(*vals);
		*vals=vals[1];
		vals[1]=NULL;
	};

	return lgba_i_addmodchain(LDAP_MOD_REPLACE, sf_strdup(attr), vals);
}

int
rmattr(const char *attr, const char *value) {

	return rmattr2(attr, value?charize(value):NULL);
};

int
rmattr2(const char *attr, char **values) {
	int z;
	if(!attr) {
		free_values(values);
		return lgba_errno=LGBA_COMMAND;
	}
	if(ecq(attr, lgba_i_node->OA) || ecq2(attr, "objectclass")) {
		free_values(values);
		return lgba_errno=LGBA_COMMAND;
	}

	if(!canIwrite(attr)) {
		free_values(values);
		return lgba_errno=LGBA_ACCESS;
	}

	z = lgba_internal_destroymodchain_s(&loaded, 1);
	lgba_i_addmodchain(LDAP_MOD_DELETE, sf_strdup(attr), values);
	lgba_internal_destroymodchain_s(&loaded, 1);
	return lgba_errno=z;
};

slist *rmtreelist=NULL;

int
lgba_internal_rmtreefunc(int i) {
	_lgba_internal_warn("+%s", loaded.dn);
	sadd(rmtreelist, loaded.dn);
	lgba_internal_rmtree(loaded.dn);

	return LGBA_NOERR;
};

int
lgba_internal_rmtree(char *dn) {
	static char *attrlist[]={ "objectclass", NULL };

	dn=(char *)sf_strdup(dn);
	lgba_search(dn, "(|(!(objectclass=*))(objectclass=*))", attrlist, NULL,NULL, 0, lgba_internal_rmtreefunc, 0);
	free(dn);

	return LGBA_NOERR;
};

int
rmobject(const char *un) {
	char *rmname=(char *)un;
	char **iatl=lgba_internal_atl;
	char *atl[]={ lgba_i_node->OA, "objectclass", NULL };
	int ec;

	lgba_internal_atl=atl;
	if(un && load(un)) { lgba_internal_atl=iatl; return lgba_errno; };
	lgba_internal_atl=iatl;

	if(!loaded.dn) return lgba_errno=LGBA_VALUE;

	if(!canIdelete(lgba_i_node->OA)) return lgba_errno=LGBA_PERM;

	if(!ld) return lgba_errno=LGBA_COMMAND;

	ldap_uncache_entry(ld, loaded.dn);

	rmname=sf_strdup(loaded.object);
	if((ec=ldap_delete_s(ld, loaded.dn))) {
		switch(ec) {
			case LDAP_SERVER_DOWN:
				if(lgba_open()) {
					lgba_free();
					free(rmname);
					return lgba_errno;
				};
				rmobject(rmname);
				free(rmname);
				return lgba_errno;
			case 66:
				{
/* CHILDREN */
static int entr=0;
if(entr) { lgba_save(1); free(rmname); return lgba_errno=LGBA_PERM; };
entr++;
if(rmtreelist) sclear(rmtreelist);
else rmtreelist=sinit();
if(!(ec=lgba_internal_rmtree(loaded.dn))) {
	while(rmtreelist->count && !ldap_delete_s(ld,
	rmtreelist->list[rmtreelist->count - 1])) {
		_lgba_internal_warn("$INFO: rm'd child '%s'",
			rmtreelist->list[rmtreelist->count - 1]);
			sdel(rmtreelist, rmtreelist->count - 1);
	};
	if(rmtreelist->count)
		_lgba_internal_warn("$INFO: failed to rm child '%s'",
	rmtreelist->list[rmtreelist->count - 1]);
	ec=rmobject(rmname);
};
entr--;
lgba_free(1);
free(rmname);
return lgba_errno;
};
			default:
				_lgba_internal_warn("-ERR: LDAP: %d/%s", ec, ldap_err2string(ec));
		};
		lgba_free();
		free(rmname);
		return lgba_errno=LGBA_LDAP;
	};

	lgba_free();
	free(rmname);
	return LGBA_NOERR;
};

int
addclass(const char *cls) {
	char **dp;
	char **vp;
	acl *ac;
	slist *sl;
	slist *dl;

	if(!loaded.count) return lgba_errno=LGBA_NOTFOUND;
	if(scfind(lgba_i_classnames, cls) == -1) return lgba_errno=LGBA_NOCLASS;
	if(scfind(loaded.objectclass, cls) != -1) return lgba_errno=LGBA_ALEX;

	/* Can I add this class to an existing class list? */
	if(!canIwrite("objectclass")) return lgba_errno=LGBA_PERM;

	/* Can I create this class? */
	if(lgba_i_getaccess(cls, "objectclass", AC_CREATE) != 1)
		return lgba_errno=LGBA_PERM;

	/* Copy classlist and add */
	sadd(loaded.objectclass, (char *)cls);

	if(!(ac=lgba_i_getcreatelist(cls)))
		return lgba_errno=LGBA_PERM;

	lgba_i_addmodchain(LDAP_MOD_REPLACE,
		sf_strdup("objectclass"),
		mkarray(loaded.objectclass, 0));

	sl=ac->attrlist; vp=sarray(sl, 0);
	dl=ac->defaults; dp=sarray(dl, 0);
	while(*vp) {
		if(ecq(*vp, lgba_i_node->OA) || ecq2(*vp, "objectclass")) {
			vp++; dp++;
			continue;
		};
		lgba_i_addmodchain(LDAP_MOD_REPLACE, sf_strdup(*vp), charize(*dp));
		vp++; dp++;
	}

	return load(loaded.object);
};

int
rmclass(const char *cls) {
	slist *sl, *dl;
	char **vp, **dp;
	acl *ac;

	if(!loaded.count) return lgba_errno=LGBA_NOTFOUND;
	if(scfind(lgba_i_classnames, cls) == -1) return lgba_errno=LGBA_NOCLASS;
	if(scfind(loaded.objectclass, cls) == -1) return lgba_errno=LGBA_NOCLASS;

	/* Can I modify the existing class list? */
	if(!canIwrite("objectclass")) return lgba_errno=LGBA_PERM;

	if(!(ac=lgba_i_getdeletelist(cls))) return lgba_errno=LGBA_PERM;

	/* Can I delete this class? */
	if(lgba_i_getaccess(cls, "objectclass", AC_DELETE) != 1) return lgba_errno=LGBA_PERM;

	/* Can't remove the last class */
	if(loaded.objectclass->count <= 1)
		return lgba_errno=LGBA_NOCLASS;

	lgba_i_addmodchain(LDAP_MOD_DELETE, sf_strdup("objectclass"), charize(cls));
	lgba_internal_destroymodchain_s(&loaded, 1);

	sl=ac->attrlist; vp=sarray(sl, 0);
	dl=ac->defaults; dp=sarray(dl, 0);
	while(*vp) {
		if(ecq(*vp, lgba_i_node->OA) || ecq(*vp, "objectclass")) {
			vp++; dp++; continue; };
		/* Delete a particular value or whole attribute */
		lgba_i_addmodchain(LDAP_MOD_DELETE, sf_strdup(*vp), **dp?charize(*dp):NULL);
		lgba_internal_destroymodchain_s(&loaded, 1);
		vp++; dp++;
	};

	return load(loaded.object);
};

char *lgba_objectname() { return loaded.object; }
char *lgba_objectdn() { return loaded.dn; }
char **lgba_objectclass() { return sarray(loaded.objectclass, 0); }

int
lgba_setuser(const char *u) {
#ifndef	LGBA_GID
#define	LGBA_GID	84
#endif
	int ngrp;
	gid_t *grp;
	char *c;

	if(getgid() == LGBA_GID || getegid() == LGBA_GID)
		goto ok;

	ngrp = getgroups(0, NULL);
	if(!ngrp)
		return LGBA_PERM;

	grp = (gid_t *)alloca(ngrp * sizeof(gid_t));
	if(!grp)
		return lgba_errno = LGBA_RSHORT;

	ngrp = getgroups(ngrp, grp);
	if(ngrp > 0) {
		while(ngrp--) {
			if(grp[ngrp] == LGBA_GID)
				goto ok;
		}
	}

	return lgba_errno=LGBA_PERM;

ok:
	if(lgba_i_user)
		free(lgba_i_user);
	lgba_i_user=sf_strdup(u);
	if(!lgba_i_user)
		return LGBA_RSHORT;

	for(c = lgba_i_user; *c; c++)
		*c = tolower(*c);

	return LGBA_NOERR;
};

void lgba_perror(const char *msg) {
	if(msg)
		_lgba_internal_warn("%s: %s", msg, lgba_err2string(lgba_errno));
	else
		_lgba_internal_warn("%s", lgba_err2string(lgba_errno));
};

int
lgba_cd(const char *s) {
	char **nvp;
	char **vp;

	if(!lgba_i_node) return lgba_errno=LGBA_CONF;

	lgba_dn_dirty=1;

	if(!s) {
		free(lgba_i_node->cur_dn);
		lgba_i_node->cur_dn=sf_strdup("");
		return LGBA_NOERR;
	};

	if(!(nvp=lgba_explode_dn((char *)s))) return lgba_errno=LGBA_VALUE;

	/* single value */
	vp=nvp-1;
	while(*(++vp))
		lgba_cd_single(*vp);

	free_values(nvp);

	return LGBA_NOERR;
};

int lgba_cd_single(char *s) {
	char *z;
	if(!s) return LGBA_VALUE;

	if(!*s) return LGBA_NOERR;

	if(eq(s, "..")) {
		char **vp=ldap_explode_dn(lgba_i_node->cur_dn, 0);
		if(!vp || !*vp) {
			free_values(vp);
			lgba_i_node->cur_dn[0]=0;
			return LGBA_VALUE;
		};
		free(lgba_i_node->cur_dn);
		lgba_i_node->cur_dn=sf_strdup(join(vp+1, ", "));
		free_values(vp);
		return LGBA_NOERR;
	};

	z=(char *)sf_malloc(strlen(lgba_i_node->OA) +
					strlen(lgba_i_node->cur_dn) +
					strlen(s) + 2);

	sprintf(z, "%s=%s%s%s", lgba_i_node->OA, s, *lgba_i_node->cur_dn?", ":"", lgba_i_node->cur_dn);

	free(lgba_i_node->cur_dn);
	lgba_i_node->cur_dn=z;

	return LGBA_NOERR;
};

char *li_cdn=NULL;

char *lgba_current_dn() {
	if(li_cdn && !lgba_dn_dirty) return li_cdn;

	li_cdn=(char *)sf_realloc(li_cdn, strlen(lgba_i_node->basedn) +
		strlen(lgba_i_node->cur_dn) + 4);

	sprintf(li_cdn, "%s%s%s",
		lgba_i_node->cur_dn,
		*lgba_i_node->cur_dn?", ":"",
		lgba_i_node->basedn
		);

	lgba_dn_dirty=0;

	return li_cdn;
};

char *lgba_internal_gendn(const char *s) {
	char *z;
	if(!s || !*s) return sf_strdup(lgba_current_dn());

	z=(char *)sf_malloc(strlen(lgba_i_node->OA) + strlen(lgba_current_dn()) + strlen(lgba_i_node->basedn) + strlen(lgba_i_node->cur_dn) + strlen(s) + 8);

	sprintf(z, "%s=%s, %s", lgba_i_node->OA, s, lgba_current_dn());

	return z;
};

int lgba_verifypwd(const char *pwd) {
	int les;
	if(!ld) return lgba_errno=LGBA_COMMAND;
	if(!loaded.dn) return lgba_errno=LGBA_NOTFOUND;

	if(!pwd || !*pwd) return LGBA_ACCESS;

	les=ldap_simple_bind_s(ld, loaded.dn, (char *)pwd);

	if(les==LDAP_SERVER_DOWN) {
		if(lgba_open()) return lgba_errno;
		return lgba_verifypwd(pwd);
	};

	les=les?LGBA_ACCESS:LGBA_NOERR;

	lgba_chnode(lgba_i_node_head->name);

	return lgba_errno=les;
};

char **lgba_objectattributes() {
	static slist *sl=NULL;
	int cnt;
	avt **ap;

	if(!loaded.dn) return NULL;
	if(!sl) sl=sinit();
	else sclear(sl);

	cnt=loaded.count;
	ap=loaded.list;
	while(cnt--) {
		sadd(sl, (*ap)->key);
		ap++;
	};

	return sarray(sl, 0);
};


