#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define __USE_XOPEN
#include <sys/types.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_CRYPT_H
#include <crypt.h>
#endif

#include <pwd.h>

#include "intern.h"

unsigned char itoa64[] =         /* 0 ... 63 => ascii - 64 */
	"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

void lgba_i_to64(char *s, long v, int n) {
        while (--n >= 0) { *s++ = itoa64[v&0x3f]; v >>= 6; }
}

char *
lgba_i_DEScrypt(const char *pwd) {
	struct timeval tv;
	char salt[10];

#ifdef	HAVE_SRANDOMDEV
	srandomdev();
#else
	srandom(time(NULL));
#endif

#ifdef	HAVE_GETTIMEOFDAY
	gettimeofday(&tv,0);
#else
	tv.tv_sec = random();
	tv.tv_usec = random();
#endif

	lgba_i_to64(&salt[0], random(), 3);
	lgba_i_to64(&salt[3], tv.tv_usec, 3);
	lgba_i_to64(&salt[6], tv.tv_sec, 2);
	salt[8] = '\0';
	return crypt(pwd, salt);
};

#ifndef _PASSWORD_LEN
#define _PASSWORD_LEN	128
#endif

char *lgba_i_makepassword(const char *en, int docrypt) {
	static char pwdbuf[_PASSWORD_LEN+1];
	char *p=NULL;

	if(!en) return NULL;

	if(*en == '{') {
		strncpy(pwdbuf, en, _PASSWORD_LEN);
		return pwdbuf;
	};

	if(!*en) return "{CRYPT}*";

	if(!strcmp(en, "-") && isatty(0) && isatty(1) && isatty(2)) {
		p=getpass("New password:");
		if(!*p) return "{CRYPT}*";
	};

	if(!strcmp(en, "--") && isatty(0) && isatty(1) && isatty(2)) {
                int mistakes=4;
                while(mistakes--) {
                        p=getpass("New password:");
                        strcpy(pwdbuf, p);
                        p=getpass("Retype new password:");
                        if(!strcmp(p, pwdbuf)) {
                                _lgba_internal_warn("+OK: Password changed!");
                                mistakes++;
                                break;
			} else if(mistakes)
				_lgba_internal_warn("Mismatch: try again.");
            };
            if(mistakes <= 0) {
				_lgba_internal_warn("-ERR: Too many tries.");
				return NULL;
            };
		if(!*p) return "{CRYPT}*";
        };

	if(!docrypt) {
		strncpy(pwdbuf, p?p:en, _PASSWORD_LEN);
	} else {
		strncpy(pwdbuf, "{CRYPT}", 7);
		strcpy(pwdbuf+7, lgba_i_DEScrypt(p?p:en));
	};

	return pwdbuf;
};


