#ifndef __READCF_H__
#define __READCF_H__

#include <sys/types.h>
#include <stdlib.h>
#include <regex.h>
#include "strfunc.h"

#define AC_CREATE	1
#define AC_DELETE	2
#define AC_READ		3
#define AC_WRITE	4

/* ======================== */
/*      attribute types     */
/* ======================== */
typedef struct _typestr {
    int tp;
    slist *attr;
    void *other;
    struct _typestr *next;
} typestr;

extern typestr *lgba_i_tphead;

/* ======================== */
/*       access classes     */
/* ======================== */
typedef struct _acl {
    slist	*users;
    slist	*attrlist;
    slist	*defaults;
    struct _acl *next;
} acl;

/* ======================== */
/*       class access       */
/* ======================== */
typedef struct _oca {
    char *classname;
    acl *create;
    acl *delete;
    acl *read;
    acl *write;
    struct _oca *next;
} oca;

extern oca *lgba_i_ocahead;

/* ======================== */
/*        validation        */
/* ======================== */
typedef struct _validate {
    slist *attrs;
    slist *exprs;
    slist *REs;
    struct _validate *next;
} Validate;

extern Validate *lgba_i_val_head;

/* ======================== */
/*        overriding        */
/* ======================== */
typedef struct _override {
	char *attr;
	char *to;
	char **(*func)(char **, char *);
    struct _override *next;
} Override;

extern Override *lgba_i_ovr_head;

typedef struct {
    char *key;
    slist *vals;
	int dequoted;
} avt;


typedef struct {
	int	protected;

	char	*object;
	char	*dn;
	slist	*objectclass;

	int		count;
	avt		**list;

	/* Modifications */
	slist	*lm_list;

	/* for iteration */
	avt		**lfa;
} answer;

extern answer loaded;

typedef struct _leafnode {
    char *name;
    slist *bindserver;
    char *binddn;
    char *bindpwd;
    char *basedn;
    char *cur_dn;
    char *defaultclass;
	char *OA;	/* object attribute */
    char *inc_attr;
    int  inc_start;
    int  inc_end;
    char *mnt_creator;
    char *mnt_created;
    char *mnt_modifier;
    char *mnt_modified;
    struct _leafnode *next;
} leafnode; 

extern leafnode *lgba_i_node_head;
extern leafnode *lgba_i_node;

typedef struct _file {
	char *name;
	char *node;
	char *filename;
	char *program;
	int append;
	char *head;
	char *body;
	char *tail;
	char *filter;
	char *sortby;
	char **attrlist;
	int mode;
	int owner;
	int group;
	struct _file *next;
} file;

extern file *lgba_i_file_head;
extern slist *lgba_i_files;

extern char *lgba_i_defaultnode;
extern char *lgba_i_defaultlang;
extern int  lgba_i_smartsave;
extern char *lgba_i_ldapserver;
extern char *lgba_i_user;

extern char *lgba_i_listformat;
extern int lgba_i_htmlmode;
extern slist *lgba_i_classnames;
extern char *lgba_i_OA;	/* object attribute */

void lgba_i_addtype(int, slist *, void *other);
void lgba_i_cfgaddclass(char *, acl *, acl *, acl *, acl *);
int lgba_i_addacl(acl **, slist *, slist *, slist *);

int lgba_i_getaccess(const char *, const char *, int);
int lgba_i_denied(char **, const char *, int);
int lgba_i_findaclreg(acl *, const char *);
oca *lgba_i_findclass(const char *);
int lgba_i_validate(const char *, const char *);
int lgba_i_validate2(const char *, const char **);
int lgba_readavfile(const char *fname, slist **, slist **);

char **lgba_i_httpize(char **, char *);
char **lgba_i_nbspize(char **, char *);
char **lgba_i_imagize(char **, char *);
char **lgba_i_timeize(char **, char *);
char **lgba_i_mapize(char **, char *);

extern int lgba_dn_dirty;

int lgba_readcf(char *);

#endif
