#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <lber.h>
#include <ldap.h>
#include <string.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include "lgba.h"
#include "friendly.h"
#include "readcf.h"
#include "pwd.h"
#include "file.h"
#include "getlists.h"
#include "intern.h"

extern answer loaded;
extern int lgba_errno;

extern LDAP *ld;

static answer	**_global_answers=NULL;

void
_lgba_free_global_answers() {
	answer **ans;

	if(!_global_answers) return;

	for(ans = _global_answers; *ans; ans++)
		lgba_free_s(*ans);

	free(_global_answers);
};

int
lgba_internal_search(char *base, char *filter, char **attrlist, int (*cntf)(int), int (*func)(int), int flags, void *sortby, int (*cmp)(const char *a, const char *b), int sortcomplex) {
	char *at;
	LDAPMessage *res;
	LDAPMessage *en;
	char **atlCopy=NULL;
	int countreturn=-1;
	int realcount=0;
	answer **ans;
	answer **saved_answers;
	int k;

	lgba_save(1);

	/* Perform a new search */
  if(!(flags & 0x02)) {

	if(!filter) filter="(|(!(objectclass=*))(objectclass=*))";

	at=(base && *base)?base:lgba_current_dn();

	if(attrlist) {
		copy_values(attrlist, &atlCopy);
		attrlist=atlCopy;
		if(cfind(atlCopy, lgba_i_node->OA) == -1)
			atlCopy[count_values(atlCopy)]=sf_strdup(lgba_i_node->OA);
		if(cfind(atlCopy, "objectclass") == -1)
			atlCopy[count_values(atlCopy)]=sf_strdup("objectclass");
	};

here:

	/* If connection is broken */
	if(!ld && lgba_open()) {
		free_values(atlCopy);
		return 0;
	};

	if((k=ldap_search_s(ld, at,
		(flags&0x01)?LDAP_SCOPE_SUBTREE:LDAP_SCOPE_ONELEVEL,
			filter, attrlist, func?0:1, &res))) {
		if(k == LDAP_SERVER_DOWN) {
			if((k=lgba_open())) return LGBA_NOERR;
			goto here;
		};
		free_values(atlCopy);
		return LGBA_NOERR;
	};
	free_values(atlCopy);

	realcount = ldap_count_entries(ld, res);
  } else {
	register answer **ans=_global_answers;
	if(!ans)
		return 0;
	realcount = 0;
	while(*ans++)
		realcount++;
  };

	if(cntf) {
		/* Save answers */
		saved_answers = _global_answers;
		_global_answers = NULL;

		countreturn=cntf(realcount);

		if(_global_answers)
			_lgba_free_global_answers();

		/* Restore answers */
		_global_answers = saved_answers;

		if(!countreturn) {
			if(!(flags & 0x02))
				ldap_msgfree(res);
			return 0;
		};
	};
	if(realcount <= 0) {
		if(!(flags & 0x02))
			ldap_msgfree(res);
		return 0;
	};

	if(!func) {
		if(!(flags & 0x02))
			ldap_msgfree(res);
		return realcount;
	};

	/* Perform a new search */
  if(!(flags & 0x02)) {

	/* Clear */
	k = 0;
	if( (ans=_global_answers) ) {
		for(; *ans; ans++, k++) {
			lgba_free_s(*ans);
			if(k >= realcount) {
				free(*ans); *ans=NULL;
			};
		};
	};

	_global_answers = (answer **)sf_realloc(_global_answers,
		(realcount+1) * sizeof(answer *));

	for(; k < realcount; k++)
		_global_answers[k] = NULL;

	k = 0;

	en=ldap_first_entry(ld, res);
	while(en) {
		if(!_global_answers[k])
			_global_answers[k] = (answer *)sf_calloc(1, sizeof(answer));

		lgba_i_fillanswer(_global_answers[k], en);
		en=ldap_next_entry(ld, en);

		if(++k > realcount)
			break;
	};
	ldap_msgfree(res);
	realcount = k;
	_global_answers[k] = NULL;

  };

	if(sortby) {
		if(sortcomplex)
			lgba_sort_entries2(_global_answers, realcount, sortby);
		else
			lgba_sort_entries(_global_answers, realcount, sortby, cmp);
	};

	if(countreturn < 0)
		countreturn = realcount;

	/* Save answer against recursive invocations */
	saved_answers = _global_answers;
	_global_answers = NULL;

	for(lgba_i_counter = k = 0; k < countreturn; k++) {

		memcpy(&loaded, saved_answers[k], sizeof(answer));
		loaded.protected = 1;

		if(func(++lgba_i_counter))
			goto qt;

		/* Destroy changes */
		lgba_internal_destroymodchain_s(&loaded, 0);

	};

	lgba_free();

qt:

	if(_global_answers)
		_lgba_free_global_answers();

	_global_answers = saved_answers;

	return lgba_i_counter;
}

int
lgba_search(char *base, char *filter, char **attrlist, char *sortby, int (*cmp)(const char *a, const char *b), int (*cntf)(int), int (*func)(int), int subtree) {
	return lgba_internal_search(base, filter, attrlist, cntf, func, subtree, sortby, cmp, 0);
};

int
lgba_search2(char *base, char *filter, char **attrlist, struct SortBy **sortby, int (*cntf)(int), int (*func)(int), int subtree) {
	return lgba_internal_search(base, filter, attrlist, cntf, func, subtree, sortby, NULL, 1);
};


