#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <ctype.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#include <strfunc.h>

#include "lgba.h"
#include "intern.h"

struct entrything {
	char	**et_vals;
	answer	*et_ans;
};

static int	(*lgba_et_cmp_fn)( const char *a, const char *b );
static int	lgba_et_cmp (const void *aa, const void *bb);

int
lgba_srt(const char *a, const char *b) {
    register int va=atoi(a);
    register int vb=atoi(b);
    if(va != vb) return va-vb;

#ifdef	HAVE_STRCOLL
    return strcoll(a, b);
#else
	return strcmp(a, b);
#endif
}

static int
lgba_et_cmp(
	const void	*aa,
	const void	*bb
)
{
	int			i, rc;
	const struct entrything	*a = (const struct entrything *)aa;
	const struct entrything	*b = (const struct entrything *)bb;

	if ( a->et_vals == NULL && b->et_vals == NULL )
		return( 0 );
	if ( a->et_vals == NULL )
		return( -1 );
	if ( b->et_vals == NULL )
		return( 1 );

	for ( i = 0; a->et_vals[i] && b->et_vals[i]; i++ ) {
		if ( (rc = (*lgba_et_cmp_fn)( a->et_vals[i], b->et_vals[i] ))
		    != 0 ) {
			return( rc );
		}
	}

	if ( a->et_vals[i] == NULL && b->et_vals[i] == NULL )
		return( 0 );
	if ( a->et_vals[i] == NULL )
		return( -1 );
	return( 1 );
}

int
lgba_sort_entries(
	answer **answers, int count,
    char	*attr,		/* NULL => sort by DN */
    int		(*cmp) (const char *, const char *)
)
{
	int			i;
	struct entrything	*et;
	char **as=NULL, **as_head=NULL;

	if(count < 0) return -1;
	if(count < 2) return 0;

	if(!(et=(struct entrything *)sf_malloc(count*sizeof(struct entrything))))
		return -1;

	if(attr)
		as=as_head=lgba_explode_dn(attr);

	if(!attr || !*attr) {
		for ( i = 0; i < count; i++ ) {
			et[i].et_ans = answers[i];
			et[i].et_vals = lgba_explode_dn( answers[i]->dn );
		};
	} else
		for ( i = 0; i < count; i++ ) {
			et[i].et_ans = answers[i];
			et[i].et_vals = sarray(getattr2_s(answers[i], *as), 0);
		}

	lgba_et_cmp_fn = cmp?cmp:lgba_srt;
	qsort(et, count, sizeof(struct entrything), lgba_et_cmp);

	/* Do a inside-attribute sorting */
	if(attr)
	while(*(++as)) {
		int itmp=0;
		for(i = 0; i < count - 1; i++) {
			if(lgba_et_cmp(et+i, et+i+1)) {
				if(i - itmp) {
					et[i].et_vals = sarray(getattr2_s(et[i].et_ans, *as), 0);
					qsort(et + itmp, i - itmp + 1,
						sizeof(struct entrything), lgba_et_cmp);
					itmp=i+1;
					continue;
				};
				itmp = i + 1;
			} else {
				et[i].et_vals = sarray(getattr2_s(et[i].et_ans, *as), 0);
			};
		}
	};

	/* Fill back */
	for (i = 0; i < count; i++)
		answers[i] = et[i].et_ans;

	free((char *)et);
	if(as_head)
		free_values(as_head);

	return 0;
}
